import numpy as np
cimport numpy as cnp
cimport cython
import os.path
import logging

from dipy.data import get_sphere
from dipy.core.sphere import disperse_charges, Sphere, HemiSphere
from tempfile import gettempdir
from libc.math cimport sqrt, exp, fabs, cos, sin, tan, acos, atan2
from math import ceil

logger = logging.getLogger(__name__)


cdef class EnhancementKernel:
    cdef double D33
    cdef double D44
    cdef double t
    cdef int kernelsize
    cdef double kernelmax
    cdef double [:, :] orientations_list
    cdef double [:, :, :, :, :] lookuptable
    cdef object sphere

    def __init__(self, D33, D44, t, force_recompute=False,
                 orientations=None, verbose=True):
        """ Compute a look-up table for the contextual
        enhancement kernel

        See :footcite:p:`Meesters2016a`, :footcite:p:`Duits2011`,
        :footcite:p:`Portegies2015a` and :footcite:p:`Portegies2015b` for
        further details about the method.

        Parameters
        ----------
        D33 : float
            Spatial diffusion
        D44 : float
            Angular diffusion
        t : float
            Diffusion time
        force_recompute : boolean, optional
            Always compute the look-up table even if it is available
            in cache.
        orientations : integer or Sphere object, optional
            Specify the number of orientations to be used with
            electrostatic repulsion, or provide a Sphere object.
            The default sphere is 'repulsion100'.
        verbose : boolean, optional
            Enable verbose mode.

        References
        ----------
        .. footbibliography::
        """

        # save parameters as class members
        self.D33 = D33
        self.D44 = D44
        self.t = t

        # define a sphere
        rng = np.random.default_rng()

        if isinstance(orientations, Sphere):
            # use the sphere defined by the user
            sphere = orientations
        elif isinstance(orientations, (int, float)):
            # electrostatic repulsion based on number of orientations
            n_pts = int(orientations)
            if n_pts == 0:
                sphere = None
            else:
                theta = np.pi * rng.random(n_pts)
                phi = 2 * np.pi * rng.random(n_pts)
                hsph_initial = HemiSphere(theta=theta, phi=phi)
                sphere, potential = disperse_charges(hsph_initial, 5000)
        else:
            # use default
            sphere = get_sphere(name="repulsion100")

        if sphere is not None:
            self.orientations_list = sphere.vertices
            self.sphere = sphere
        else:
            self.orientations_list = np.zeros((0,0))
            self.sphere = None

        # file location of the lut table for saving/loading
        kernellutpath = os.path.join(gettempdir(),
                                     "kernel_d33@%4.2f_d44@%4.2f_t@%4.2f_numverts%d.npy" \
                                       % (D33, D44, t, len(self.orientations_list)))

        # if LUT exists, load
        if not force_recompute and os.path.isfile(kernellutpath):
            if verbose:
                logger.info("The kernel already exists. Loading from " + kernellutpath)
            self.lookuptable = np.load(kernellutpath)

        # else, create
        else:
            if verbose:
                logger.info("The kernel doesn't exist yet. Computing...")
            self.create_lookup_table(verbose)
            if self.sphere is not None:
                np.save(kernellutpath, self.lookuptable)

    def get_lookup_table(self):
        """ Return the computed look-up table.
        """
        return self.lookuptable

    def get_orientations(self):
        """ Return the orientations.
        """
        return self.orientations_list

    def get_sphere(self):
        """ Get the sphere corresponding with the orientations
        """
        return self.sphere

    def evaluate_kernel(self, x, y, r, v):
        """ Evaluate the kernel at position x relative to
        position y, with orientation r relative to orientation v.

        Parameters
        ----------
        x : 1D ndarray
            Position x
        y : 1D ndarray
            Position y
        r : 1D ndarray
            Orientation r
        v : 1D ndarray
            Orientation v

        Returns
        -------
        kernel_value : double
        """
        return self.k2(x, y, r, v)

    @cython.wraparound(False)
    @cython.boundscheck(False)
    @cython.nonecheck(False)
    @cython.cdivision(True)
    cdef void create_lookup_table(self, verbose=True):
        """ Compute the look-up table based on the parameters set
        during class initialization

        Parameters
        ----------
        verbose : boolean
            Enable verbose mode.
        """
        self.estimate_kernel_size(verbose)

        cdef:
            double [:, :] orientations = np.copy(self.orientations_list)
            cnp.npy_intp OR1 = orientations.shape[0]
            cnp.npy_intp OR2 = orientations.shape[0]
            cnp.npy_intp N = self.kernelsize
            cnp.npy_intp hn = (N-1)/2
            cnp.npy_intp angv, angr, xp, yp, zp
            double [:] x
            double [:] y
            cdef double [:, :, :, :, :] lookuptablelocal
            double kmax = self.kernelmax
            double l1norm
            double kernelval

        lookuptablelocal = np.zeros((OR1, OR2, N, N, N))
        x = np.zeros(3)
        y = np.zeros(3) # constant at (0,0,0)

        with nogil:

            for angv in range(OR1):
                for angr in range(OR2):
                    for xp in range(-hn, hn + 1):
                        for yp in range(-hn, hn + 1):
                            for zp in range(-hn, hn + 1):

                                x[0] = xp
                                x[1] = yp
                                x[2] = zp

                                lookuptablelocal[angv,
                                                 angr,
                                                 xp + hn,
                                                 yp + hn,
                                                 zp + hn] = self.k2(x, y, orientations[angr,:], orientations[angv,:])

        # save to class member
        self.lookuptable = lookuptablelocal

    @cython.wraparound(False)
    @cython.boundscheck(False)
    @cython.nonecheck(False)
    @cython.cdivision(True)
    cdef void estimate_kernel_size(self, verbose=True):
        """ Estimates the dimensions the kernel should
        have based on the kernel parameters.

        Parameters
        ----------
        verbose : boolean
            Enable verbose mode.
        """

        cdef:
            double [:] x
            double [:] y
            double [:] r
            double [:] v
            double i

        x = np.array([0., 0., 0.])
        y = np.array([0., 0., 0.])
        r = np.array([0., 0., 1.])
        v = np.array([0., 0., 1.])

        # evaluate at origin
        self.kernelmax = self.k2(x, y, r, v)

        with nogil:

            # determine a good kernel size
            i = 0.0
            while True:
                i += 0.1
                x[2] = i
                kval = self.k2(x, y, r, v) / self.kernelmax
                if kval < 0.1:
                    break

        N = ceil(i) * 2
        if N % 2 == 0:
            N -= 1

        if verbose:
            logger.info("Dimensions of kernel: %dx%dx%d" % (N, N, N))

        self.kernelsize = N

    @cython.wraparound(False)
    @cython.boundscheck(False)
    @cython.nonecheck(False)
    cdef double k2(self, double [:] x, double [:] y,
                   double [:] r, double [:] v) noexcept nogil:
        """ Evaluate the kernel at position x relative to
        position y, with orientation r relative to orientation v.

        Parameters
        ----------
        x : 1D ndarray
            Position x
        y : 1D ndarray
            Position y
        r : 1D ndarray
            Orientation r
        v : 1D ndarray
            Orientation v

        Returns
        -------
        kernel_value : double
        """
        cdef:
            double [:] a
            double [:,:] transm
            double [:] arg1
            double [:] arg2p
            double [:] arg2
            double [:] c
            double kernelval

        with gil:

            a = np.subtract(x, y)
            transm = np.transpose(R(euler_angles(v)))
            arg1 = np.dot(transm, a)
            arg2p = np.dot(transm, r)
            arg2 = euler_angles(arg2p)
            c = self.coordinate_map(arg1[0], arg1[1], arg1[2],
                                    arg2[0], arg2[1])
            kernelval = self.kernel(c)

        return kernelval

    @cython.wraparound(False)
    @cython.boundscheck(False)
    @cython.nonecheck(False)
    @cython.cdivision(True)
    cdef double [:] coordinate_map(self, double x, double y,
                                   double z, double beta,
                                   double gamma) noexcept nogil:
        """ Compute a coordinate map for the kernel

        Parameters
        ----------
        x : double
            X position
        y : double
            Y position
        z : double
            Z position
        beta : double
            First Euler angle
        gamma : double
            Second Euler angle

        Returns
        -------
        c : 1D ndarray
            array of coordinates for kernel
        """

        cdef:
            double [:] c
            double q
            double cg
            double cotq2

        with gil:

            c = np.zeros(6)

        if beta == 0:
            c[0] = x
            c[1] = y
            c[2] = z
            c[3] = c[4] = c[5] = 0

        else:
            q = fabs(beta)
            cg = cos(gamma)
            sg = sin(gamma)
            cotq2 = 1.0 / tan(q/2)

            c[0] = -0.5*z*beta*cg + \
                    x*(1 - (beta*beta*cg*cg * (1 - 0.5*q*cotq2)) / (q*q)) - \
                    (y*beta*beta*cg*sg * (1 - 0.5*q*cotq2)) / (q*q)
            c[1] = -0.5*z*beta*sg - \
                    (x*beta*beta*cg*sg * (1 - 0.5*q*cotq2)) / (q*q) + \
                    y * (1 - (beta*beta*sg*sg * (1 - 0.5*q*cotq2)) / (q*q))
            c[2] = 0.5*x*beta*cg + 0.5*y*beta*sg + \
                   z * (1 + ((1 - 0.5*q*cotq2) * (-beta*beta*cg*cg - \
                        beta*beta*sg*sg)) / (q*q))
            c[3] = beta * (-sg)
            c[4] = beta * cg
            c[5] = 0

        return c

    @cython.wraparound(False)
    @cython.boundscheck(False)
    @cython.nonecheck(False)
    @cython.cdivision(True)
    cdef double kernel(self, double [:] c) noexcept nogil:
        """ Internal function, evaluates the kernel based on the coordinate map.

        Parameters
        ----------
        c : 1D ndarray
            array of coordinates for kernel

        Returns
        -------
        kernel_value : double
        """
        cdef double output = 1 / (8*sqrt(2))
        output *= sqrt(PI)*self.t*sqrt(self.t*self.D33)*sqrt(self.D33*self.D44)
        output *= 1 / (16*PI*PI*self.D33*self.D33*self.D44*self.D44*self.t*self.t*self.t*self.t)
        output *= exp(-sqrt((c[0]*c[0] + c[1]*c[1]) / (self.D33*self.D44) + \
                   (c[2]*c[2] / self.D33 + (c[3]*c[3]+c[4]*c[4]) / self.D44) * \
                   (c[2]*c[2] / self.D33 + (c[3]*c[3]+c[4]*c[4]) / self.D44) + \
                    c[5]*c[5]/self.D44) / (4*self.t))
        return output

cdef double PI = 3.1415926535897932

@cython.wraparound(False)
@cython.boundscheck(False)
cdef double [:] euler_angles(double [:] inp) noexcept nogil:
    """ Compute the Euler angles for a given input vector

    Parameters
    ----------
    inp : 1D ndarray
        Input vector

    Returns
    -------
    euler_angles : 1D ndarray
    """
    cdef:
        double x
        double y
        double z
        double [:] output

    x = inp[0]
    y = inp[1]
    z = inp[2]

    with gil:

        output = np.zeros(3)

    # handle the case (0,0,1)
    if x*x < 10e-6 and y*y < 10e-6 and (z-1) * (z-1) < 10e-6:
        output[0] = 0
        output[1] = 0

    # handle the case (0,0,-1)
    elif x*x < 10e-6 and y*y < 10e-6 and (z+1) * (z+1) < 10e-6:
        output[0] = PI
        output[1] = 0

    # all other cases
    else:
        output[0] = acos(z)
        output[1] = atan2(y, x)

    return output

@cython.wraparound(False)
@cython.boundscheck(False)
cdef double [:,:] R(double [:] inp) noexcept nogil:
    """ Compute the Rotation matrix for a given input vector

    Parameters
    ----------
    inp : 1D ndarray
        Input vector

    Returns
    -------
    rotation_matrix : 2D ndarray
    """
    cdef:
        double beta
        double gamma
        double [:] output
        double cb
        double sb
        double cg
        double sg

    beta = inp[0]
    gamma = inp[1]

    with gil:

        output = np.zeros(9)

    cb = cos(beta)
    sb = sin(beta)
    cg = cos(gamma)
    sg = sin(gamma)

    output[0] = cb * cg
    output[1] = -sg
    output[2] = cg * sb
    output[3] = cb * sg
    output[4] = cg
    output[5] = sb * sg
    output[6] = -sb
    output[7] = 0
    output[8] = cb

    with gil:

        return np.reshape(output, (3,3))



from numbers import Number

import numpy as np

from dipy.denoise.nlmeans_block import nlmeans_block
from dipy.testing.decorators import warning_for_keywords


@warning_for_keywords()
def non_local_means(
    arr, sigma, *, mask=None, patch_radius=1, block_radius=5, rician=True
):
    r"""Non-local means for denoising 3D and 4D images, using blockwise
    averaging approach.

    See :footcite:p:`Coupe2008` and :footcite:p:`Coupe2012` for further details
    about the method.

    Parameters
    ----------
    arr : 3D or 4D ndarray
        The array to be denoised
    mask : 3D ndarray
        Mask on data where the non-local means will be applied.
    sigma : float or ndarray
        standard deviation of the noise estimated from the data
    patch_radius : int
        patch size is ``2 x patch_radius + 1``. Default is 1.
    block_radius : int
        block size is ``2 x block_radius + 1``. Default is 5.
    rician : boolean
        If True the noise is estimated as Rician, otherwise Gaussian noise
        is assumed.

    Returns
    -------
    denoised_arr : ndarray
        the denoised ``arr`` which has the same shape as ``arr``.

    References
    ----------
    .. footbibliography::

    """
    if isinstance(sigma, np.ndarray) and sigma.size == 1:
        sigma = sigma.item()
    if isinstance(sigma, np.ndarray):
        if arr.ndim == 3:
            raise ValueError("sigma should be a scalar for 3D data", sigma)
        if not np.issubdtype(sigma.dtype, np.number):
            raise ValueError("sigma should be an array of floats", sigma)
        if arr.ndim == 4 and sigma.ndim != 1:
            raise ValueError("sigma should be a 1D array for 4D data", sigma)
        if arr.ndim == 4 and sigma.shape[0] != arr.shape[-1]:
            raise ValueError(
                "sigma should have the same length as the last "
                "dimension of arr for 4D data",
                sigma,
            )
    else:
        if not isinstance(sigma, Number):
            raise ValueError("sigma should be a float", sigma)
        # if sigma is a scalar and arr is 4D, we assume the same sigma for all
        if arr.ndim == 4:
            sigma = np.array([sigma] * arr.shape[-1])

    if mask is None and arr.ndim > 2:
        mask = np.ones((arr.shape[0], arr.shape[1], arr.shape[2]), dtype="f8")
    else:
        mask = np.ascontiguousarray(mask, dtype="f8")

    if mask.ndim != 3:
        raise ValueError("mask needs to be a 3D ndarray", mask.shape)

    if arr.ndim == 3:
        return np.array(
            nlmeans_block(
                np.double(arr), mask, patch_radius, block_radius, sigma, int(rician)
            )
        ).astype(arr.dtype)
    elif arr.ndim == 4:
        denoised_arr = np.zeros_like(arr)
        for i in range(arr.shape[-1]):
            denoised_arr[..., i] = np.array(
                nlmeans_block(
                    np.double(arr[..., i]),
                    mask,
                    patch_radius,
                    block_radius,
                    sigma[i],
                    int(rician),
                )
            ).astype(arr.dtype)

        return denoised_arr

    else:
        raise ValueError("Only 3D or 4D array are supported!", arr.shape)



from os.path import join as pjoin
from tempfile import TemporaryDirectory

import nibabel as nib
import numpy as np
import numpy.testing as npt

from dipy.align.streamlinear import BundleMinDistanceMetric
from dipy.data import get_fnames
from dipy.io.image import load_nifti_data
from dipy.io.stateful_tractogram import Space, StatefulTractogram
from dipy.io.streamline import load_tractogram, save_tractogram
from dipy.segment.mask import median_otsu
from dipy.segment.tests.test_mrf import create_image
from dipy.tracking.streamline import Streamlines, set_number_of_points
from dipy.utils.optpkg import optional_package
from dipy.workflows.segment import (
    ClassifyTissueFlow,
    LabelsBundlesFlow,
    MedianOtsuFlow,
    RecoBundlesFlow,
)

sklearn, has_sklearn, _ = optional_package("sklearn")


def test_median_otsu_flow():
    with TemporaryDirectory() as out_dir:
        data_path, _, _ = get_fnames(name="small_25")
        volume = load_nifti_data(data_path)
        save_masked = True
        median_radius = 3
        numpass = 3
        autocrop = False
        vol_idx = "0,"
        dilate = 0
        finalize_mask = False

        mo_flow = MedianOtsuFlow()
        mo_flow.run(
            data_path,
            out_dir=out_dir,
            save_masked=save_masked,
            median_radius=median_radius,
            numpass=numpass,
            autocrop=autocrop,
            vol_idx=vol_idx,
            dilate=dilate,
            finalize_mask=finalize_mask,
        )

        mask_name = mo_flow.last_generated_outputs["out_mask"]
        masked_name = mo_flow.last_generated_outputs["out_masked"]

        vol_idx = [
            0,
        ]
        masked, mask = median_otsu(
            volume,
            vol_idx=vol_idx,
            median_radius=median_radius,
            numpass=numpass,
            autocrop=autocrop,
            dilate=dilate,
        )

        result_mask_data = load_nifti_data(pjoin(out_dir, mask_name))
        npt.assert_array_equal(result_mask_data.astype(np.uint8), mask)

        result_masked = nib.load(pjoin(out_dir, masked_name))
        result_masked_data = np.asanyarray(result_masked.dataobj)

        npt.assert_array_equal(np.round(result_masked_data), masked)


def test_recobundles_flow():
    with TemporaryDirectory() as out_dir:
        data_path = get_fnames(name="fornix")

        fornix = load_tractogram(data_path, "same", bbox_valid_check=False).streamlines

        f = Streamlines(fornix)
        f1 = f.copy()

        f2 = f1[:15].copy()
        f2._data += np.array([40, 0, 0])

        f.extend(f2)

        f2_path = pjoin(out_dir, "f2.trk")
        sft = StatefulTractogram(f2, data_path, Space.RASMM)
        save_tractogram(sft, f2_path, bbox_valid_check=False)

        f1_path = pjoin(out_dir, "f1.trk")
        sft = StatefulTractogram(f, data_path, Space.RASMM)
        save_tractogram(sft, f1_path, bbox_valid_check=False)

        rb_flow = RecoBundlesFlow(force=True)
        rb_flow.run(
            f1_path,
            f2_path,
            greater_than=0,
            clust_thr=10,
            model_clust_thr=5.0,
            reduction_thr=10,
            out_dir=out_dir,
        )

        labels = rb_flow.last_generated_outputs["out_recognized_labels"]
        recog_trk = rb_flow.last_generated_outputs["out_recognized_transf"]

        rec_bundle = load_tractogram(
            recog_trk, "same", bbox_valid_check=False
        ).streamlines
        npt.assert_equal(len(rec_bundle) == len(f2), True)

        label_flow = LabelsBundlesFlow(force=True)
        label_flow.run(f1_path, labels, out_dir=out_dir)

        recog_bundle = label_flow.last_generated_outputs["out_bundle"]
        rec_bundle_org = load_tractogram(
            recog_bundle, "same", bbox_valid_check=False
        ).streamlines

        BMD = BundleMinDistanceMetric()
        nb_pts = 20
        static = set_number_of_points(f2, nb_points=nb_pts)
        moving = set_number_of_points(rec_bundle_org, nb_points=nb_pts)

        BMD.setup(static, moving)
        x0 = np.array([0, 0, 0, 0, 0, 0, 1.0, 1.0, 1, 0, 0, 0])  # affine
        bmd_value = BMD.distance(x0.tolist())

        npt.assert_equal(bmd_value < 1, True)


def test_classify_tissue_flow():
    with TemporaryDirectory() as out_dir:
        data = create_image()
        data_path = pjoin(out_dir, "data.nii.gz")
        nib.save(nib.Nifti1Image(data, np.eye(4)), data_path)

        args = {
            "input_files": data_path,
            "method": "hmrf",
            "nclass": 4,
            "beta": 0.1,
            "tolerance": 0.0001,
            "max_iter": 10,
            "out_dir": out_dir,
        }
        flow = ClassifyTissueFlow()
        flow.run(**args)

        tissue = flow.last_generated_outputs["out_tissue"]
        pve = flow.last_generated_outputs["out_pve"]

        tissue_data = load_nifti_data(tissue)
        pve_data = load_nifti_data(pve)

        npt.assert_equal(tissue_data.shape, data.shape)
        npt.assert_equal(tissue_data.max(), 4)
        npt.assert_equal(tissue_data.min(), 0)

        npt.assert_equal(pve_data.shape, (data.shape) + (4,))
        npt.assert_equal(pve_data.max(), 1)

        npt.assert_raises(SystemExit, flow.run, data_path)
        npt.assert_raises(SystemExit, flow.run, data_path, method="random")
        npt.assert_raises(SystemExit, flow.run, data_path, method="dam")
        npt.assert_raises(SystemExit, flow.run, data_path, method="hmrf")

    if has_sklearn:
        with TemporaryDirectory() as out_dir:
            data = np.random.rand(3, 3, 3, 7) * 100  # Simulated random data
            bvals = np.array([0, 100, 500, 1000, 1500, 2000, 3000])
            data_path = pjoin(out_dir, "data.nii.gz")
            bvals_path = pjoin(out_dir, "bvals")
            np.savetxt(bvals_path, bvals)
            nib.save(nib.Nifti1Image(data, np.eye(4)), data_path)

            args = {
                "input_files": data_path,
                "bvals_file": bvals_path,
                "method": "dam",
                "wm_threshold": 0.5,
                "out_dir": out_dir,
            }
            flow = ClassifyTissueFlow()
            flow.run(**args)

            tissue = flow.last_generated_outputs["out_tissue"]
            pve = flow.last_generated_outputs["out_pve"]

            tissue_data = load_nifti_data(tissue)
            pve_data = load_nifti_data(pve)

            npt.assert_equal(tissue_data.shape, data.shape[:-1])
            npt.assert_equal(tissue_data.max(), 2)
            npt.assert_equal(tissue_data.min(), 0)

            npt.assert_equal(pve_data.shape, (data.shape[:-1]) + (2,))
            npt.assert_equal(pve_data.max(), 1)



"""
=================================================
Using Various Stopping Criterion for Tractography
=================================================
The stopping criterion determines if the tracking stops or continues at each
tracking position. The tracking stops when it reaches an ending region
(e.g. low FA, gray matter or corticospinal fluid regions) or exits the image
boundaries. The tracking also stops if the direction getter has no direction
to follow.

Each stopping criterion determines if the stopping is 'valid' or
'invalid'. A streamline is 'valid' when the stopping criterion determines if
the streamline stops in a position classified as 'ENDPOINT' or 'OUTSIDEIMAGE'.
A streamline is 'invalid' when it stops in a position classified as
'TRACKPOINT' or 'INVALIDPOINT'. These conditions are described below. The
'LocalTracking' generator can be set to output all generated streamlines
or only the 'valid' ones. See :footcite:t:`Girard2014` and
:footcite:p:`Smith2012` for more details on these methods.

This example is an extension of the
:ref:`sphx_glr_examples_built_fiber_tracking_tracking_deterministic.py`
example. We begin by loading the data, creating a seeding mask from white
matter voxels of the corpus callosum, fitting a Constrained Spherical
Deconvolution (CSD) reconstruction model and creating the maximum
deterministic direction getter.
"""

import matplotlib.pyplot as plt
import numpy as np

from dipy.core.gradients import gradient_table
from dipy.data import default_sphere, get_fnames
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, load_nifti_data
from dipy.io.stateful_tractogram import Space, StatefulTractogram
from dipy.io.streamline import save_trk
from dipy.reconst.csdeconv import ConstrainedSphericalDeconvModel, auto_response_ssst
from dipy.reconst.dti import TensorModel, fractional_anisotropy
from dipy.tracking import utils
from dipy.tracking.stopping_criterion import (
    ActStoppingCriterion,
    BinaryStoppingCriterion,
    ThresholdStoppingCriterion,
)
from dipy.tracking.streamline import Streamlines
from dipy.tracking.tracker import deterministic_tracking
from dipy.viz import actor, colormap, has_fury, window

# Enables/disables interactive visualization
interactive = False

hardi_fname, hardi_bval_fname, hardi_bvec_fname = get_fnames(name="stanford_hardi")
label_fname = get_fnames(name="stanford_labels")
_, _, f_pve_wm = get_fnames(name="stanford_pve_maps")

data, affine, hardi_img = load_nifti(hardi_fname, return_img=True)
labels = load_nifti_data(label_fname)
bvals, bvecs = read_bvals_bvecs(hardi_bval_fname, hardi_bvec_fname)
gtab = gradient_table(bvals, bvecs=bvecs)

white_matter = load_nifti_data(f_pve_wm)

seed_mask = labels == 2
seed_mask[white_matter < 0.5] = 0
seeds = utils.seeds_from_mask(seed_mask, affine, density=2)

response, ratio = auto_response_ssst(gtab, data, roi_radii=10, fa_thr=0.7)
csd_model = ConstrainedSphericalDeconvModel(gtab, response)
csd_fit = csd_model.fit(data, mask=white_matter)

###############################################################################
# Threshold Stopping Criterion
# ============================
# A scalar map can be used to define where the tracking stops. The threshold
# stopping criterion uses a scalar map to stop the tracking whenever the
# interpolated scalar value is lower than a fixed threshold. Here, we show
# an example using the fractional anisotropy (FA) map of the DTI model.
# The threshold stopping criterion uses a trilinear interpolation at the
# tracking position.
#
# **Parameters**
#
# - metric_map: numpy array [:, :, :]
# - threshold: float
#
# **Stopping States**
#
# - 'ENDPOINT': stops at a position where metric_map < threshold; the
#   streamline reached the target stopping area.
# - 'OUTSIDEIMAGE': stops at a position outside of metric_map; the streamline
#   reached an area outside the image where no direction data is available.
# - 'TRACKPOINT': stops at a position because no direction is available; the
#   streamline is stopping where metric_map >= threshold, but there is no valid
#   direction to follow.
# - 'INVALIDPOINT': N/A.
#

tensor_model = TensorModel(gtab)
tenfit = tensor_model.fit(data, mask=labels > 0)
FA = fractional_anisotropy(tenfit.evals)

threshold_criterion = ThresholdStoppingCriterion(FA, 0.2)

fig = plt.figure()
mask_fa = FA.copy()
mask_fa[mask_fa < 0.2] = 0
plt.xticks([])
plt.yticks([])
plt.imshow(
    mask_fa[:, :, data.shape[2] // 2].T,
    cmap="gray",
    origin="lower",
    interpolation="nearest",
)
fig.tight_layout()
fig.savefig("threshold_fa.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Thresholded fractional anisotropy map.

streamline_generator = deterministic_tracking(
    seeds,
    threshold_criterion,
    affine,
    step_size=0.5,
    return_all=True,
    sh=csd_fit.shm_coeff,
    max_angle=30.0,
    sphere=default_sphere,
)
streamlines = Streamlines(streamline_generator)
sft = StatefulTractogram(streamlines, hardi_img, Space.RASMM)
save_trk(sft, "tractogram_probabilistic_thresh_all.trk")

if has_fury:
    scene = window.Scene()
    scene.add(actor.line(streamlines, colors=colormap.line_colors(streamlines)))
    window.record(
        scene=scene, out_path="tractogram_deterministic_thresh_all.png", size=(800, 800)
    )
    if interactive:
        window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
#  Corpus Callosum using deterministic tractography with a thresholded
#  fractional anisotropy mask.
#
#
#
# Binary Stopping Criterion
# =========================
# A binary mask can be used to define where the tracking stops. The binary
# stopping criterion stops the tracking whenever the tracking position is
# outside the mask. Here, we show how to obtain the binary stopping criterion
# from the white matter mask defined above. The binary stopping criterion uses
# a nearest-neighborhood interpolation at the tracking position.
#
# **Parameters**
#
# - mask: numpy array [:, :, :]
#
# **Stopping States**
#
# - 'ENDPOINT': stops at a position where mask = 0; the streamline
#   reached the target stopping area.
# - 'OUTSIDEIMAGE': stops at a position outside of metric_map; the streamline
#   reached an area outside the image where no direction data is available.
# - 'TRACKPOINT': stops at a position because no direction is available; the
#   streamline is stopping where mask > 0, but there is no valid direction to
#   follow.
# - 'INVALIDPOINT': N/A.
#

binary_criterion = BinaryStoppingCriterion(white_matter == 1)

fig = plt.figure()
plt.xticks([])
plt.yticks([])
fig.tight_layout()
plt.imshow(
    white_matter[:, :, data.shape[2] // 2].T,
    cmap="gray",
    origin="lower",
    interpolation="nearest",
)

fig.savefig("white_matter_mask.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# White matter binary mask.

streamline_generator = deterministic_tracking(
    seeds,
    binary_criterion,
    affine,
    step_size=0.5,
    return_all=True,
    sh=csd_fit.shm_coeff,
    max_angle=30.0,
    sphere=default_sphere,
)
streamlines = Streamlines(streamline_generator)
sft = StatefulTractogram(streamlines, hardi_img, Space.RASMM)
save_trk(sft, "tractogram_deterministic_binary_all.trk")

if has_fury:
    scene = window.Scene()
    scene.add(actor.line(streamlines, colors=colormap.line_colors(streamlines)))
    window.record(
        scene=scene, out_path="tractogram_deterministic_binary_all.png", size=(800, 800)
    )
    if interactive:
        window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Corpus Callosum using deterministic tractography with a binary white
#  matter mask.
#
#
#
# ACT Stopping Criterion
# ======================
# Anatomically-constrained tractography (ACT) :footcite:p:`Smith2012` uses
# information from anatomical images to determine when the tractography stops.
# The ``include_map`` defines when the streamline reached a 'valid' stopping
# region (e.g. gray matter partial volume estimation (PVE) map) and the
# ``exclude_map`` defines when the streamline reached an 'invalid' stopping
# region (e.g. corticospinal fluid PVE map). The background of the anatomical
# image should be added to the ``include_map`` to keep streamlines exiting
# the brain (e.g. through the brain stem). The ACT stopping criterion uses
# a trilinear interpolation at the tracking position.
#
# **Parameters**
#
# - ``include_map``: numpy array ``[:, :, :]``,
# - ``exclude_map``: numpy array ``[:, :, :]``,
#
# **Stopping States**
#
# - 'ENDPOINT': stops at a position where ``include_map`` > 0.5; the streamline
#   reached the target stopping area.
# - 'OUTSIDEIMAGE': stops at a position outside of ``include_map`` or
#   ``exclude_map``; the streamline reached an area outside the image where no
#   direction data is available.
# - 'TRACKPOINT': stops at a position because no direction is available; the
#   streamline is stopping where ``include_map`` < 0.5 and
#   ``exclude_map`` < 0.5, but there is no valid direction to follow.
# - 'INVALIDPOINT': ``exclude_map`` > 0.5; the streamline reach a position
#   which is anatomically not plausible.
#

f_pve_csf, f_pve_gm, f_pve_wm = get_fnames(name="stanford_pve_maps")
pve_csf_data = load_nifti_data(f_pve_csf)
pve_gm_data = load_nifti_data(f_pve_gm)
pve_wm_data = load_nifti_data(f_pve_wm)

background = np.ones(pve_gm_data.shape)
background[(pve_gm_data + pve_wm_data + pve_csf_data) > 0] = 0

include_map = pve_gm_data
include_map[background > 0] = 1
exclude_map = pve_csf_data

act_criterion = ActStoppingCriterion(include_map, exclude_map)

fig = plt.figure()
plt.subplot(121)
plt.xticks([])
plt.yticks([])
plt.imshow(
    include_map[:, :, data.shape[2] // 2].T,
    cmap="gray",
    origin="lower",
    interpolation="nearest",
)

plt.subplot(122)
plt.xticks([])
plt.yticks([])
plt.imshow(
    exclude_map[:, :, data.shape[2] // 2].T,
    cmap="gray",
    origin="lower",
    interpolation="nearest",
)

fig.tight_layout()
fig.savefig("act_maps.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Include (left) and exclude (right) maps for ACT.

streamline_generator = deterministic_tracking(
    seeds,
    act_criterion,
    affine,
    step_size=0.5,
    return_all=True,
    sh=csd_fit.shm_coeff,
    max_angle=30.0,
    sphere=default_sphere,
)
streamlines = Streamlines(streamline_generator)
sft = StatefulTractogram(streamlines, hardi_img, Space.RASMM)
save_trk(sft, "tractogram_deterministic_act_all.trk")

if has_fury:
    scene = window.Scene()
    scene.add(actor.line(streamlines, colors=colormap.line_colors(streamlines)))
    window.record(
        scene=scene, out_path="tractogram_deterministic_act_all.png", size=(800, 800)
    )
    if interactive:
        window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Corpus Callosum using deterministic tractography with ACT stopping
#  criterion.

streamline_generator = deterministic_tracking(
    seeds,
    act_criterion,
    affine,
    step_size=0.5,
    return_all=False,
    sh=csd_fit.shm_coeff,
    max_angle=30.0,
    sphere=default_sphere,
)
streamlines = Streamlines(streamline_generator)
sft = StatefulTractogram(streamlines, hardi_img, Space.RASMM)
save_trk(sft, "tractogram_deterministic_act_valid.trk")

if has_fury:
    scene = window.Scene()
    scene.add(actor.line(streamlines, colors=colormap.line_colors(streamlines)))
    window.record(
        scene=scene, out_path="tractogram_deterministic_act_valid.png", size=(800, 800)
    )
    if interactive:
        window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Corpus Callosum using deterministic tractography with ACT stopping
# criterion. Streamlines ending in gray matter region only.
#
#
#
# The threshold and binary stopping criterion use respectively a scalar map
# and a binary mask to stop the tracking. The ACT stopping criterion use
# partial volume fraction (PVE) maps from an anatomical image to stop the
# tracking. Additionally, the ACT stopping criterion determines if the
# tracking stopped in expected regions (e.g. gray matter) and allows the
# user to get only streamlines stopping in those regions.
#
# Notes
# -----
# Currently,the proposed method that cuts streamlines going through
# subcortical gray matter regions is not implemented. The
# backtracking technique for streamlines reaching INVALIDPOINT is not
# implemented either :footcite:p:`Smith2012`.
#
#
# References
# ----------
#
# .. footbibliography::
#



"""
================================
Creating a new combined workflow
================================

A ``CombinedWorkflow`` is a series of DIPY_ workflows organized together in a
way that the output of a workflow serves as input for the next one.

First create your ``CombinedWorkflow`` class. Your ``CombinedWorkflow`` class
file is usually located in the ``dipy/workflows`` directory.
"""

import os

from dipy.workflows.combined_workflow import CombinedWorkflow

###############################################################################
# ``CombinedWorkflow`` is the base class that will be extended to create our
# combined workflow.
from dipy.workflows.denoise import NLMeansFlow
from dipy.workflows.segment import MedianOtsuFlow

###############################################################################
# ``MedianOtsuFlow`` and ``NLMeansFlow`` will be combined to create our
# processing section.


class DenoiseAndSegment(CombinedWorkflow):
    """
    ``DenoiseAndSegment`` is the name of our combined workflow. Note that
    it needs to extend CombinedWorkflow for everything to work properly.
    """

    def _get_sub_flows(self):
        return [NLMeansFlow, MedianOtsuFlow]

        """
        It is mandatory to implement this method if you want to make all the
        sub workflows parameters available in commandline.
        """

    def run(
        self,
        input_files,
        out_dir="",
        out_denoised="processed.nii.gz",
        out_mask="brain_mask.nii.gz",
        out_masked="dwi_masked.nii.gz",
    ):
        """
        Parameters
        ----------
        input_files : string
            Path to the input files. This path may contain wildcards to
            process multiple inputs at once.

        out_dir : string, optional
            Where the resulting file will be saved. (default '')

        out_denoised : string, optional
            Name of the denoised file to be saved.

        out_mask : string, optional
            Name of the Otsu mask file to be saved.

        out_masked : string, optional
            Name of the Otsu masked file to be saved.
        """

        """
        Just like a normal workflow, it is mandatory to have out_dir as a
        parameter. It is also mandatory to put 'out_' in front of every
        parameter that is going to be an output. Lastly, all out_ params needs
        to be at the end of the params list.

        The class docstring part is very important, you need to document
        every parameter as they will be used with inspection to build the
        command line argument parser.
        """

        io_it = self.get_io_iterator()

        for fnames in io_it:
            in_fname = fnames[0]
            _out_denoised = os.path.basename(fnames[1])
            _out_mask = os.path.basename(fnames[2])
            _out_masked = os.path.basename(fnames[3])

            nl_flow = NLMeansFlow()
            self.run_sub_flow(
                nl_flow, in_fname, out_dir=out_dir, out_denoised=_out_denoised
            )
            denoised = nl_flow.last_generated_outputs["out_denoised"]

            me_flow = MedianOtsuFlow()
            self.run_sub_flow(
                me_flow,
                denoised,
                out_dir=out_dir,
                out_mask=_out_mask,
                out_masked=_out_masked,
            )


###############################################################################
# Use ``self.get_io_iterator()`` in every workflow you create. This creates
# an ``IOIterator`` object that create output file names and directory
# structure based on the inputs and some other advanced output strategy
# parameters.
#
# Iterating on the ``IOIterator`` object you created previously you
# conveniently get all input and output paths for every input file
# found when globbin the input parameters.
#
# In the ``IOIterator`` loop you can see how we create a new ``NLMeans``
# workflow then run it using ``self.run_sub_flow``. Running it this way will
# pass any workflow specific parameter that was retrieved from the command line
# and will append the ones you specify as optional parameters (``out_dir``
# in this case).
#
# Lastly, the outputs paths are retrieved using
# ``workflow.last_generated_outputs``. This allows to use ``denoise`` as the
# input for the ``MedianOtsuFlow``.
#
#
# This is it for the combined workflow class! Now to be able to call it easily
# via command line, you need to add this workflow in 2 different files:
#
# - ``<dipy_root>/pyproject.toml``: open this file and add the following line
#   to the ``[project.scripts]`` section:
#   ``dipy_denoise_segment = "dipy.workflows.cli:run"``
#
# - ``<dipy_root>/dipy/workflows/cli.py``: open this file and add the workflow
#   information to the ``cli_flows`` dictionary. The key is the name of the
#   command line command and the value is a tuple with the module name and the
#   workflow class name. In this case it would be:
#   ``"dipy_denoise_segment": ("dipy.workflows.my_combined_workflow",
#   "DenoiseAndSegment")``
#
# That`s it! Now you can call your workflow from anywhere with the command line.
# Let's just call the script you just made with ``-h`` to see the argparser help
# text::
#
#    dipy_denoise_segment --help
#
# You should see all your parameters available along with some extra common
# ones like logging file and force overwrite. Also all the documentation you
# wrote about each parameter is there.



from dipy.segment.cythonutils cimport Data2D, Shape
cimport numpy as cnp

cdef class Feature:
    cdef int is_order_invariant

    cdef Shape c_infer_shape(Feature self, Data2D datum) noexcept nogil
    cdef void c_extract(Feature self, Data2D datum, Data2D out) noexcept nogil

    cpdef infer_shape(Feature self, datum)
    cpdef extract(Feature self, datum)


cdef class CythonFeature(Feature):
    pass

# The IdentityFeature class returns the datum as-is. This is useful for metric
# that does not require any pre-processing.
cdef class IdentityFeature(CythonFeature):
    pass

# The ResampleFeature class returns the datum resampled. This is useful for
# metric like SumPointwiseEuclideanMetric that does require a consistent
# number of points between datum.
cdef class ResampleFeature(CythonFeature):
    cdef cnp.npy_intp nb_points



"""Function for determining the effective number of processes to be used."""

from multiprocessing import cpu_count
from warnings import warn


def determine_num_processes(num_processes):
    """Determine the effective number of processes for parallelization.

    - For `num_processes = None`` return the maximum number of cores retrieved
    by cpu_count().

    - For ``num_processes > 0``, return this value.

    - For ``num_processes < 0``, return the maximal number of cores minus
    ``num_processes + 1``. In particular ``num_processes = -1`` will use as
    many cores as possible.

    - For ``num_processes = 0`` a ValueError is raised.

    Parameters
    ----------
    num_processes : int or None
        Desired number of processes to be used.
    """
    if not isinstance(num_processes, int) and num_processes is not None:
        raise TypeError("num_processes must be an int or None")

    if num_processes == 0:
        raise ValueError("num_processes cannot be 0")

    try:
        if num_processes is None:
            return cpu_count()

        if num_processes < 0:
            return max(1, cpu_count() + num_processes + 1)
    except NotImplementedError:
        warn("Cannot determine number of cores. Using only 1.", stacklevel=2)
        return 1

    return num_processes



"""
========================================
Visualize bundles and metrics on bundles
========================================

First, let's download some available datasets. Here we are using a dataset
which provides metrics and bundles.
"""

import numpy as np

from dipy.data import fetch_bundles_2_subjects, read_bundles_2_subjects
from dipy.tracking.streamline import length, transform_streamlines
from dipy.viz import actor, window

fetch_bundles_2_subjects()
dix = read_bundles_2_subjects(
    subj_id="subj_1", metrics=["fa"], bundles=["cg.left", "cst.right"]
)

###############################################################################
# Store fractional anisotropy.

fa = dix["fa"]

###############################################################################
# Store grid to world transformation matrix.

affine = dix["affine"]

###############################################################################
# Store the cingulum bundle. A bundle is a list of streamlines.

bundle = dix["cg.left"]

###############################################################################
# It happened that this bundle is in world coordinates and therefore we need
# to transform it into native image coordinates so that it is in the same
# coordinate space as the ``fa`` image.

bundle_native = transform_streamlines(bundle, np.linalg.inv(affine))

###############################################################################
# Show every streamline with an orientation color
# ===============================================
#
# This is the default option when you are using ``line`` or ``streamtube``.

scene = window.Scene()

stream_actor = actor.line(bundle_native)

scene.set_camera(
    position=(-176.42, 118.52, 128.20),
    focal_point=(113.30, 128.31, 76.56),
    view_up=(0.18, 0.00, 0.98),
)

scene.add(stream_actor)

# Uncomment the line below to show to display the window
# window.show(scene, size=(600, 600), reset_camera=False)
window.record(scene=scene, out_path="bundle1.png", size=(600, 600))

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# One orientation color for every streamline.
#
#
# You may wonder how we knew how to set the camera. This is very easy. You just
# need to run ``window.show`` once to see how you want to see the object and
# then close the window and call the ``camera_info`` method which prints the
# position, focal point and view up vectors of the camera.

scene.camera_info()

###############################################################################
# Show every point with a value from a volume with default colormap
# =================================================================
#
# Here we will need to input the ``fa`` map in ``streamtube`` or ``line``.

scene.clear()
stream_actor2 = actor.line(bundle_native, colors=fa, linewidth=0.1)

###############################################################################
# We can also show the scalar bar.

bar = actor.scalar_bar()

scene.add(stream_actor2)
scene.add(bar)

# window.show(scene, size=(600, 600), reset_camera=False)
window.record(scene=scene, out_path="bundle2.png", size=(600, 600))

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Every point with a color from FA.
#
#
# Show every point with a value from a volume with your colormap
# ==============================================================
#
# Here we will need to input the ``fa`` map in ``streamtube``

scene.clear()

hue = (0.0, 0.0)  # red only
saturation = (0.0, 1.0)  # white to red

lut_cmap = actor.colormap_lookup_table(hue_range=hue, saturation_range=saturation)

stream_actor3 = actor.line(
    bundle_native, colors=fa, linewidth=0.1, lookup_colormap=lut_cmap
)
bar2 = actor.scalar_bar(lookup_table=lut_cmap)

scene.add(stream_actor3)
scene.add(bar2)

# window.show(scene, size=(600, 600), reset_camera=False)
window.record(scene=scene, out_path="bundle3.png", size=(600, 600))

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Every point with a color from FA using a non default colormap.
#
#
# Show every bundle with a specific color
# ========================================
#
# You can have a bundle with a specific color. In this example, we are choosing
# orange.

scene.clear()
stream_actor4 = actor.line(bundle_native, colors=(1.0, 0.5, 0), linewidth=0.1)

scene.add(stream_actor4)

# window.show(scene, size=(600, 600), reset_camera=False)
window.record(scene=scene, out_path="bundle4.png", size=(600, 600))

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Entire bundle with a specific color.
#
#
# Show every streamline of a bundle with a different color
# ========================================================
#
# Let's make a colormap where every streamline of the bundle is colored by
# its length.

scene.clear()

lengths = length(bundle_native)

hue = (0.5, 0.5)  # blue only
saturation = (0.0, 1.0)  # black to white

lut_cmap = actor.colormap_lookup_table(
    scale_range=(lengths.min(), lengths.max()),
    hue_range=hue,
    saturation_range=saturation,
)

stream_actor5 = actor.line(
    bundle_native, colors=lengths, linewidth=0.1, lookup_colormap=lut_cmap
)

scene.add(stream_actor5)
bar3 = actor.scalar_bar(lookup_table=lut_cmap)

scene.add(bar3)

# window.show(scene, size=(600, 600), reset_camera=False)
window.record(scene=scene, out_path="bundle5.png", size=(600, 600))

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Color every streamline by the length of the streamline
#
#
# Show every point of every streamline with a different color
# ============================================================
#
# In this case in which we want to have a color per point and per streamline,
# we can create a list of the colors to correspond to the list of streamlines
# (bundles). Here in ``colors`` we will insert some random RGB colors.

scene.clear()

rng = np.random.default_rng()

colors = [rng.random(streamline.shape) for streamline in bundle_native]

stream_actor6 = actor.line(
    bundle_native, colors=np.asarray(colors, dtype=object), linewidth=0.2
)

scene.add(stream_actor6)

# window.show(scene, size=(600, 600), reset_camera=False)
window.record(scene=scene, out_path="bundle6.png", size=(600, 600))

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Random colors per point per streamline.
#
#
# In summary, we showed that there are many useful ways for visualizing maps
# on bundles.



import timeit

cimport numpy as cnp
import numpy as np
from numpy.testing import (assert_, assert_almost_equal, assert_raises,
                            assert_array_equal)
from dipy.utils.fast_numpy import random, seed
from dipy.utils.fast_numpy cimport (
    cross,
    dot,
    norm,
    normalize,
    random_vector,
    random_perpendicular_vector, take,
    random_point_within_circle,
    take,
)


def test_norm():
    # Test that the norm equal numpy norm.
    cdef double[:] vec_view
    for _ in range(10):
        vec_view = vec = np.random.random(3)
        assert_almost_equal(norm(&vec_view[0]), np.linalg.norm(vec))


def test_normalize():
    # Test that the normalize vector as a norm of 1.
    cdef double[:] vec_view
    for _ in range(10):
        vec_view = vec = np.random.random(3)
        normalize(&vec_view[0])
        assert_almost_equal(np.linalg.norm(vec), 1)


def test_dot():
    # Test that dot is faster and equal to numpy.dot.
    cdef double[:] vec_view1
    cdef double[:] vec_view2
    for _ in range(10):
        vec_view1 = vec1 = np.random.random(3)
        vec_view2 = vec2 = np.random.random(3)
        assert_almost_equal(dot(&vec_view1[0], &vec_view2[0]),
            np.dot(vec1, vec2))

    vec_view1 = vec1 = np.random.random(3)
    vec_view2 = vec2 = np.random.random(3)

    def __dot():
        dot(&vec_view1[0], &vec_view2[0])

    def __npdot():
        np.dot(vec1, vec2)

    number = 100000
    time_dot = timeit.timeit(__dot, number=number)
    time_npdot = timeit.timeit(__npdot, number=number)
    assert_(time_dot < time_npdot)


# TODO: Check why this test fails
# def test_cross():
#     # Test that cross is faster and equal to numpy.cross.
#     cdef double[:] vec1
#     cdef double[:] vec2
#     cdef double[:] out
#     out = np.zeros(3, dtype=float)
#     for _ in range(10):
#         vec1 = np.random.random(3)
#         vec2 = np.random.random(3)
#         cross(&out[0], &vec1[0], &vec2[0])
#         assert_array_equal(out, np.cross(vec1, vec2))

#     vec1 = np.random.random(3)
#     vec2 = np.random.random(3)

#     def __cross():
#         cross(&out[0], &vec1[0], &vec2[0])

#     def __npcross():
#         np.cross(vec1, vec2)

#     number = 10000
#     time_cross = timeit.timeit(__cross, number=number)
#     time_npcross = timeit.timeit(__npcross, number=number)
#     assert_(time_cross < time_npcross)


def test_random_vector():
    # Test that that the random vector is of norm 1
    cdef double[:] test = np.zeros(3, dtype=float)
    for _ in range(10):
        random_vector(&test[0])
        assert_almost_equal(np.linalg.norm(test), 1)
        assert_(np.all(test >= np.double(-1)))
        assert_(np.all(test <= np.double(1)))


def test_random_perpendicular_vector():
    # Test that the random vector is of norm 1 and perpendicular
    cdef double[:] test = np.zeros(3, dtype=float)
    cdef double[:] vec = np.zeros(3, dtype=float)
    for _ in range(10):
        random_vector(&vec[0])
        random_perpendicular_vector(&test[0], &vec[0])
        assert_almost_equal(np.linalg.norm(test), 1)
        assert_(np.all(test >= np.double(-1)))
        assert_(np.all(test <= np.double(1)))
        assert_almost_equal(np.dot(vec, test), 0)


def test_random():
    # Test that random numbers are between 0 and 1 and that the mean is 0.5.
    for _ in range(10):
        vec = random()
        assert_(vec < 1 and vec > 0)

    random_number_list = []
    for _ in range(10000):
        random_number_list.append(random())
    assert_almost_equal(np.mean(random_number_list), 0.5, decimal=1)

    # Test the random generator seed input
    for s in [0, 1, np.iinfo(np.uint32).max]:
        seed(s)
    for s in [np.iinfo(np.uint32).max + 1, -1]:
        assert_raises(OverflowError, seed, s)


def test_random_point_within_circle():
    # Test that the random point is within the circle
    for r in np.arange(0, 5, 0.2):
        pts = random_point_within_circle(r)
        assert_(np.linalg.norm(pts) <= r)


def test_take():
    # Test that the take function is equal to numpy.take
    cdef int n_indices = 5
    cdef double[:] odf = np.random.random(10)
    cdef cnp.npy_intp[:] indices = np.random.randint(0, 10, n_indices, dtype=np.intp)
    cdef double[:] values_out = np.zeros(n_indices, dtype=float)
    take(&odf[0], &indices[0], n_indices, &values_out[0])
    assert_array_equal(values_out, np.take(odf, indices))


import numpy as np
import numpy.testing as npt

from dipy.core.sphere import HemiSphere, unique_edges
from dipy.data import default_sphere
from dipy.reconst.recspeed import (
    local_maxima,
    remove_similar_vertices,
    search_descending,
)


def test_local_maxima():
    sphere = default_sphere
    vertices, faces = sphere.vertices, sphere.faces
    edges = unique_edges(faces)

    # Check that the first peak is == max(odf)
    odf = abs(vertices.sum(-1))
    peak_values, peak_index = local_maxima(odf, edges)
    npt.assert_equal(max(odf), peak_values[0])
    npt.assert_equal(max(odf), odf[peak_index[0]])

    # Create an artificial odf with a few peaks
    odf = np.zeros(len(vertices))
    odf[1] = 1.0
    odf[143] = 143.0
    odf[361] = 361.0
    peak_values, peak_index = local_maxima(odf, edges)
    npt.assert_array_equal(peak_values, [361, 143, 1])
    npt.assert_array_equal(peak_index, [361, 143, 1])

    # Check that neighboring points can both be peaks
    odf = np.zeros(len(vertices))
    point1, point2 = edges[0]
    odf[[point1, point2]] = 1.0
    peak_values, peak_index = local_maxima(odf, edges)
    npt.assert_array_equal(peak_values, [1.0, 1.0])
    npt.assert_(point1 in peak_index)
    npt.assert_(point2 in peak_index)

    # Repeat with a hemisphere
    hemisphere = HemiSphere(xyz=vertices, faces=faces)
    vertices, edges = hemisphere.vertices, hemisphere.edges

    # Check that the first peak is == max(odf)
    odf = abs(vertices.sum(-1))
    peak_values, peak_index = local_maxima(odf, edges)
    npt.assert_equal(max(odf), peak_values[0])
    npt.assert_equal(max(odf), odf[peak_index[0]])

    # Create an artificial odf with a few peaks
    odf = np.zeros(len(vertices))
    odf[1] = 1.0
    odf[143] = 143.0
    odf[300] = 300.0
    peak_value, peak_index = local_maxima(odf, edges)
    npt.assert_array_equal(peak_value, [300, 143, 1])
    npt.assert_array_equal(peak_index, [300, 143, 1])

    # Check that neighboring points can both be peaks
    odf = np.zeros(len(vertices))
    point1, point2 = edges[0]
    odf[[point1, point2]] = 1.0
    peak_values, peak_index = local_maxima(odf, edges)
    npt.assert_array_equal(peak_values, [1.0, 1.0])
    npt.assert_(point1 in peak_index)
    npt.assert_(point2 in peak_index)

    # Should raise an error if odf has nans
    odf[20] = np.nan
    npt.assert_raises(ValueError, local_maxima, odf, edges)

    # Should raise an error if edge values are too large to index odf
    edges[0, 0] = 9999
    odf[20] = 0
    npt.assert_raises(IndexError, local_maxima, odf, edges)


def test_remove_similar_peaks():
    vertices = np.array(
        [
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
            [1.1, 1.0, 0.0],
            [0.0, 2.0, 1.0],
            [2.0, 1.0, 0.0],
            [1.0, 0.0, 0.0],
        ]
    )
    norms = np.sqrt((vertices * vertices).sum(-1))
    vertices = vertices / norms[:, None]

    # Return unique vertices
    uv = remove_similar_vertices(vertices, 0.01)
    npt.assert_array_equal(uv, vertices[:6])

    # Return vertices with mapping and indices
    uv, mapping, index = remove_similar_vertices(
        vertices, 0.01, return_mapping=True, return_index=True
    )
    npt.assert_array_equal(uv, vertices[:6])
    npt.assert_array_equal(mapping, list(range(6)) + [0])
    npt.assert_array_equal(index, range(6))

    # Test mapping with different angles
    uv, mapping = remove_similar_vertices(vertices, 0.01, return_mapping=True)
    npt.assert_array_equal(uv, vertices[:6])
    npt.assert_array_equal(mapping, list(range(6)) + [0])
    uv, mapping = remove_similar_vertices(vertices, 30, return_mapping=True)
    npt.assert_array_equal(uv, vertices[:4])
    npt.assert_array_equal(mapping, list(range(4)) + [1, 0, 0])
    uv, mapping = remove_similar_vertices(vertices, 60, return_mapping=True)
    npt.assert_array_equal(uv, vertices[:3])
    npt.assert_array_equal(mapping, list(range(3)) + [0, 1, 0, 0])

    # Test index with different angles
    uv, index = remove_similar_vertices(vertices, 0.01, return_index=True)
    npt.assert_array_equal(uv, vertices[:6])
    npt.assert_array_equal(index, range(6))
    uv, index = remove_similar_vertices(vertices, 30, return_index=True)
    npt.assert_array_equal(uv, vertices[:4])
    npt.assert_array_equal(index, range(4))
    uv, index = remove_similar_vertices(vertices, 60, return_index=True)
    npt.assert_array_equal(uv, vertices[:3])
    npt.assert_array_equal(index, range(3))


def test_search_descending():
    a = np.linspace(10.0, 1.0, 10)

    npt.assert_equal(search_descending(a, 1.0), 1)
    npt.assert_equal(search_descending(a, 0.89), 2)
    npt.assert_equal(search_descending(a, 0.79), 3)

    # Test small array
    npt.assert_equal(search_descending(a[:1], 1.0), 1)
    npt.assert_equal(search_descending(a[:1], 0.0), 1)
    npt.assert_equal(search_descending(a[:1], 0.5), 1)

    # Test very small array
    npt.assert_equal(search_descending(a[:0], 1.0), 0)



import types
import warnings

import numpy as np
from numpy.linalg import norm
import numpy.testing as npt
from numpy.testing import (
    assert_allclose,
    assert_almost_equal,
    assert_array_almost_equal,
    assert_array_equal,
    assert_equal,
    assert_raises,
)

from dipy.testing import assert_arrays_equal, assert_true
from dipy.testing.decorators import set_random_number_generator
from dipy.testing.memory import get_type_refcount
from dipy.tracking.streamline import (
    Streamlines,
    center_streamlines,
    cluster_confidence,
    deform_streamlines,
    orient_by_rois,
    orient_by_streamline,
    relist_streamlines,
    select_by_rois,
    select_random_set_of_streamlines,
    transform_streamlines,
    unlist_streamlines,
    values_from_volume,
)
from dipy.tracking.streamlinespeed import (
    compress_streamlines,
    length,
    set_number_of_points,
)

streamline = np.array(
    [
        [82.20181274, 91.36505890, 43.15737152],
        [82.38442230, 91.79336548, 43.87036514],
        [82.48710632, 92.27861023, 44.56298065],
        [82.53310394, 92.78545380, 45.24635315],
        [82.53793335, 93.26902008, 45.94785309],
        [82.48797607, 93.75003815, 46.64939880],
        [82.35533142, 94.25181580, 47.32533264],
        [82.15484619, 94.76634216, 47.97451019],
        [81.90982819, 95.28792572, 48.60244370],
        [81.63336945, 95.78153229, 49.23971176],
        [81.35479736, 96.24868011, 49.89558792],
        [81.08713531, 96.69807434, 50.56812668],
        [80.81504822, 97.14285278, 51.24193192],
        [80.52591705, 97.56719971, 51.92168427],
        [80.26599884, 97.98269653, 52.61848068],
        [80.04635620, 98.38131714, 53.33855820],
        [79.84691620, 98.77052307, 54.06955338],
        [79.57667542, 99.13599396, 54.78985596],
        [79.23351288, 99.43207550, 55.51065063],
        [78.84815979, 99.64141846, 56.24016571],
        [78.47383881, 99.77347565, 56.99299240],
        [78.12837219, 99.81330872, 57.76969528],
        [77.80438995, 99.85082245, 58.55574799],
        [77.49439240, 99.88065338, 59.34777069],
        [77.21414185, 99.85343933, 60.15090561],
        [76.96416473, 99.82772827, 60.96406937],
        [76.74712372, 99.80519104, 61.78676605],
        [76.52263641, 99.79122162, 62.60765076],
        [76.03757477, 100.08692169, 63.24152374],
        [75.44867706, 100.35265350, 63.79513168],
        [74.78033447, 100.57255554, 64.27278900],
        [74.11605835, 100.77330780, 64.76428986],
        [73.51222992, 100.98779297, 65.32373047],
        [72.97387695, 101.23387146, 65.93502045],
        [72.47355652, 101.49151611, 66.57343292],
        [71.99834442, 101.72480774, 67.23979950],
        [71.56909180, 101.98665619, 67.92664337],
        [71.18083191, 102.29483795, 68.61888123],
        [70.81879425, 102.63343048, 69.31127167],
        [70.47422791, 102.98672485, 70.00532532],
        [70.10092926, 103.28502655, 70.70999908],
        [69.69512177, 103.51667023, 71.42147064],
        [69.27423096, 103.71351624, 72.13452911],
        [68.91260529, 103.81676483, 72.89796448],
        [68.60788727, 103.81982422, 73.69258118],
        [68.34162903, 103.76619720, 74.49915314],
        [68.08542633, 103.70635223, 75.30856323],
        [67.83590698, 103.60187531, 76.11553955],
        [67.56822968, 103.44821930, 76.90870667],
        [67.28399658, 103.25878906, 77.68825531],
        [67.00117493, 103.03740692, 78.45989227],
        [66.72718048, 102.80329895, 79.23099518],
        [66.46197510, 102.54130554, 79.99622345],
        [66.20803833, 102.22305298, 80.74387360],
        [65.96872711, 101.88980865, 81.48987579],
        [65.72864532, 101.59316254, 82.25085449],
        [65.47808075, 101.33383942, 83.02194214],
        [65.21841431, 101.11295319, 83.80186462],
        [64.95678711, 100.94080353, 84.59326935],
        [64.71759033, 100.82022095, 85.40114594],
        [64.48053741, 100.73490143, 86.21411896],
        [64.24304199, 100.65074158, 87.02709198],
        [64.01773834, 100.55318451, 87.84204865],
        [63.83801651, 100.41996765, 88.66333008],
        [63.70982361, 100.25119019, 89.48779297],
        [63.60707855, 100.06730652, 90.31262207],
        [63.46164322, 99.91001892, 91.13648224],
        [63.26287842, 99.78648376, 91.95485687],
        [63.03713226, 99.68377686, 92.76905823],
        [62.81192398, 99.56619263, 93.58140564],
        [62.57145309, 99.42708588, 94.38592529],
        [62.32259369, 99.25592804, 95.18167114],
        [62.07497787, 99.05770111, 95.97154236],
        [61.82253647, 98.83877563, 96.75438690],
        [61.59536743, 98.59293365, 97.53706360],
        [61.46530151, 98.30503845, 98.32772827],
        [61.39904785, 97.97928619, 99.11172485],
        [61.33279419, 97.65353394, 99.89572906],
        [61.26067352, 97.30914307, 100.67123413],
        [61.19459534, 96.96743011, 101.44847107],
        [61.19580460, 96.63417053, 102.23215485],
        [61.26572037, 96.29887390, 103.01185608],
        [61.39840698, 95.96297455, 103.78307343],
        [61.57207870, 95.64262390, 104.55268097],
        [61.78163528, 95.35540771, 105.32629395],
        [62.06700134, 95.09746552, 106.08564758],
        [62.39427185, 94.85724640, 106.83369446],
        [62.74076462, 94.62278748, 107.57482147],
        [63.11461639, 94.40107727, 108.30641937],
        [63.53397751, 94.20418549, 109.02002716],
        [64.00019836, 94.03809357, 109.71183777],
        [64.43580627, 93.87523651, 110.42416382],
        [64.84857941, 93.69993591, 111.14715576],
        [65.26740265, 93.51858521, 111.86515808],
        [65.69511414, 93.36718750, 112.58474731],
        [66.10470581, 93.22719574, 113.31711578],
        [66.45891571, 93.06028748, 114.07256317],
        [66.78582001, 92.90560913, 114.84281921],
        [67.11138916, 92.79004669, 115.62040710],
        [67.44729614, 92.75711823, 116.40135193],
        [67.75688171, 92.98265076, 117.16111755],
        [68.02041626, 93.28012848, 117.91371155],
        [68.25725555, 93.53466797, 118.69052124],
        [68.46047974, 93.63263702, 119.51107788],
        [68.62039948, 93.62007141, 120.34690094],
        [68.76782227, 93.56475067, 121.18331909],
        [68.90222168, 93.46326447, 122.01765442],
        [68.99872589, 93.30039978, 122.84759521],
        [69.04119873, 93.05428314, 123.66156769],
        [69.05086517, 92.74394989, 124.45450592],
        [69.02742004, 92.40427399, 125.23509979],
        [68.95466614, 92.09059143, 126.02339935],
        [68.84975433, 91.79674530, 126.81564331],
        [68.72673798, 91.53726196, 127.61715698],
        [68.60685730, 91.30300140, 128.42681885],
        [68.50636292, 91.12481689, 129.25317383],
        [68.39311218, 91.01572418, 130.08976746],
        [68.25946808, 90.94654083, 130.92756653],
    ],
    dtype=np.float32,
)

streamline_64bit = streamline.astype(np.float64)

streamlines = [
    streamline[[0, 10]],
    streamline,
    streamline[::2],
    streamline[::3],
    streamline[::5],
    streamline[::6],
]
streamlines_64bit = [
    streamline_64bit[[0, 10]],
    streamline_64bit,
    streamline_64bit[::2],
    streamline_64bit[::3],
    streamline_64bit[::4],
    streamline_64bit[::5],
]

heterogeneous_streamlines = [
    streamline_64bit,
    streamline_64bit.reshape((-1, 6)),
    streamline_64bit.reshape((-1, 2)),
]


def length_python(xyz, along=False):
    xyz = np.asarray(xyz, dtype=np.float64)
    if xyz.shape[0] < 2:
        if along:
            return np.array([0])
        return 0
    dists = np.sqrt((np.diff(xyz, axis=0) ** 2).sum(axis=1))
    if along:
        return np.cumsum(dists)
    return np.sum(dists)


def set_number_of_points_python(xyz, n_pols=3):
    def _extrap(xyz, cumlen, distance):
        """Helper function for extrapolate"""
        ind = np.where((cumlen - distance) > 0)[0][0]
        len0 = cumlen[ind - 1]
        len1 = cumlen[ind]
        Ds = distance - len0
        Lambda = Ds / (len1 - len0)
        return Lambda * xyz[ind] + (1 - Lambda) * xyz[ind - 1]

    cumlen = np.zeros(xyz.shape[0])
    cumlen[1:] = length_python(xyz, along=True)
    step = cumlen[-1] / (n_pols - 1)

    ar = np.arange(0, cumlen[-1], step)
    if np.abs(ar[-1] - cumlen[-1]) < np.finfo("f4").eps:
        ar = ar[:-1]

    xyz2 = [_extrap(xyz, cumlen, distance) for distance in ar]
    return np.vstack((np.array(xyz2), xyz[-1]))


def test_set_number_of_points():
    # Test resampling of only one streamline
    nb_points = 12
    new_streamline_cython = set_number_of_points(streamline, nb_points=nb_points)
    new_streamline_python = set_number_of_points_python(streamline, nb_points)
    assert_equal(len(new_streamline_cython), nb_points)
    # Using a 5 digits precision because of streamline is in float32.
    assert_array_almost_equal(new_streamline_cython, new_streamline_python, 5)

    new_streamline_cython = set_number_of_points(streamline_64bit, nb_points=nb_points)
    new_streamline_python = set_number_of_points_python(streamline_64bit, nb_points)
    assert_equal(len(new_streamline_cython), nb_points)
    assert_array_almost_equal(new_streamline_cython, new_streamline_python)

    res = []
    simple_streamline = np.array([[0, 0, 0], [1, 1, 1], [2, 2, 2]], "f4")
    for nb_points in range(2, 200):
        new_streamline_cython = set_number_of_points(
            simple_streamline, nb_points=nb_points
        )
        res.append(nb_points - len(new_streamline_cython))
    assert_equal(np.sum(res), 0)

    # Test resampling of multiple streamlines of different nb_points
    nb_points = 12
    new_streamlines_cython = set_number_of_points(streamlines, nb_points=nb_points)

    for i, s in enumerate(streamlines):
        new_streamline_python = set_number_of_points_python(s, nb_points)
        # Using a 5 digits precision because of streamline is in float32.
        assert_array_almost_equal(new_streamlines_cython[i], new_streamline_python, 5)

    # ArraySequence
    arrseq = Streamlines(streamlines)
    new_streamlines_as_seq_cython = set_number_of_points(arrseq, nb_points=nb_points)
    assert_array_almost_equal(new_streamlines_as_seq_cython, new_streamlines_cython)

    new_streamlines_cython = set_number_of_points(
        streamlines_64bit, nb_points=nb_points
    )

    for i, s in enumerate(streamlines_64bit):
        new_streamline_python = set_number_of_points_python(s, nb_points)
        assert_array_almost_equal(new_streamlines_cython[i], new_streamline_python)

    # ArraySequence
    arrseq = Streamlines(streamlines_64bit)
    new_streamlines_as_seq_cython = set_number_of_points(arrseq, nb_points=nb_points)
    assert_array_almost_equal(new_streamlines_as_seq_cython, new_streamlines_cython)

    # Test streamlines with mixed dtype
    streamlines_mixed_dtype = [
        streamline,
        streamline.astype(np.float64),
        streamline.astype(np.int32),
        streamline.astype(np.int64),
    ]
    nb_points_mixed_dtype = [
        len(s)
        for s in set_number_of_points(streamlines_mixed_dtype, nb_points=nb_points)
    ]
    assert_array_equal(
        nb_points_mixed_dtype, [nb_points] * len(streamlines_mixed_dtype)
    )

    # Test streamlines with different shape
    new_streamlines_cython = set_number_of_points(
        heterogeneous_streamlines, nb_points=nb_points
    )

    for i, s in enumerate(heterogeneous_streamlines):
        new_streamline_python = set_number_of_points_python(s, nb_points)
        assert_array_almost_equal(new_streamlines_cython[i], new_streamline_python)

    # Test streamline with integer dtype
    new_streamline = set_number_of_points(streamline.astype(np.int32))
    assert_equal(new_streamline.dtype, np.float32)
    new_streamline = set_number_of_points(streamline.astype(np.int64))
    assert_equal(new_streamline.dtype, np.float64)

    # Test empty list
    assert_equal(set_number_of_points([]), [])

    # Test streamline having only one point
    assert_raises(ValueError, set_number_of_points, np.array([[1, 2, 3]]))

    # We do not support list of lists, it should be numpy ndarray.
    streamline_unsupported = [[1, 2, 3], [4, 5, 5], [2, 1, 3], [4, 2, 1]]
    assert_raises(AttributeError, set_number_of_points, streamline_unsupported)

    # Test setting number of points of a numpy with flag WRITABLE=False
    streamline_readonly = streamline.copy()
    streamline_readonly.setflags(write=False)
    assert_equal(len(set_number_of_points(streamline_readonly, nb_points=42)), 42)

    # Test setting computing length of a numpy with flag WRITABLE=False
    streamlines_readonly = []
    for s in streamlines:
        streamlines_readonly.append(s.copy())
        streamlines_readonly[-1].setflags(write=False)

    assert_equal(
        len(set_number_of_points(streamlines_readonly, nb_points=42)),
        len(streamlines_readonly),
    )

    streamlines_readonly = []
    for s in streamlines_64bit:
        streamlines_readonly.append(s.copy())
        streamlines_readonly[-1].setflags(write=False)

    assert_equal(
        len(set_number_of_points(streamlines_readonly, nb_points=42)),
        len(streamlines_readonly),
    )

    # Test if nb_points is less than 2
    assert_raises(
        ValueError,
        set_number_of_points,
        [np.ones((10, 3)), np.ones((10, 3))],
        nb_points=1,
    )


@set_random_number_generator(1234)
def test_set_number_of_points_memory_leaks(rng):
    # Test some dtypes
    dtypes = [np.float32, np.float64, np.int32, np.int64]
    for dtype in dtypes:
        s_rng = np.random.default_rng(1234)
        NB_STREAMLINES = 10000
        streamlines = [
            s_rng.standard_normal((s_rng.integers(10, 100), 3)).astype(dtype)
            for _ in range(NB_STREAMLINES)
        ]

        list_refcount_before = get_type_refcount()["list"]

        rstreamlines = set_number_of_points(streamlines, nb_points=2)
        list_refcount_after = get_type_refcount()["list"]
        del (
            rstreamlines
        )  # Delete `rstreamlines` because it holds a reference to `list  # noqa: F841

        # Calling `set_number_of_points` should increase the refcount of `list`
        #  by one since we kept the returned value.
        assert_equal(list_refcount_after, list_refcount_before + 1)

    # Test mixed dtypes
    NB_STREAMLINES = 10000
    streamlines = []
    for i in range(NB_STREAMLINES):
        dtype = dtypes[i % len(dtypes)]
        streamlines.append(
            rng.standard_normal((rng.integers(10, 100), 3)).astype(dtype)
        )

    list_refcount_before = get_type_refcount()["list"]
    rstreamlines = set_number_of_points(streamlines, nb_points=2)  # noqa: F841
    list_refcount_after = get_type_refcount()["list"]

    # Calling `set_number_of_points` should increase the refcount of `list`
    #  by one since we kept the returned value.
    assert_equal(list_refcount_after, list_refcount_before + 1)


def test_length():
    # Test length of only one streamline
    length_streamline_cython = length(streamline)
    length_streamline_python = length_python(streamline)
    assert_almost_equal(length_streamline_cython, length_streamline_python)

    length_streamline_cython = length(streamline_64bit)
    length_streamline_python = length_python(streamline_64bit)
    assert_almost_equal(length_streamline_cython, length_streamline_python)

    # Test computing length of multiple streamlines of different nb_points
    length_streamlines_cython = length(streamlines)

    for i, s in enumerate(streamlines):
        length_streamline_python = length_python(s)
        assert_array_almost_equal(
            length_streamlines_cython[i], length_streamline_python
        )

    length_streamlines_cython = length(streamlines_64bit)

    for i, s in enumerate(streamlines_64bit):
        length_streamline_python = length_python(s)
        assert_array_almost_equal(
            length_streamlines_cython[i], length_streamline_python
        )

    # ArraySequence
    # Test length of only one streamline
    length_streamline_cython = length(streamline_64bit)
    length_streamline_arrseq = length(Streamlines([streamline]))
    assert_almost_equal(length_streamline_arrseq, length_streamline_cython)

    length_streamline_cython = length(streamline_64bit)
    length_streamline_arrseq = length(Streamlines([streamline_64bit]))
    assert_almost_equal(length_streamline_arrseq, length_streamline_cython)

    # Test computing length of multiple streamlines of different nb_points
    length_streamlines_cython = length(streamlines)
    length_streamlines_arrseq = length(Streamlines(streamlines))
    assert_array_almost_equal(length_streamlines_arrseq, length_streamlines_cython)

    length_streamlines_cython = length(streamlines_64bit)
    length_streamlines_arrseq = length(Streamlines(streamlines_64bit))
    assert_array_almost_equal(length_streamlines_arrseq, length_streamlines_cython)

    # Test on a sliced ArraySequence
    length_streamlines_cython = length(streamlines_64bit[::2])
    length_streamlines_arrseq = length(Streamlines(streamlines_64bit)[::2])
    assert_array_almost_equal(length_streamlines_arrseq, length_streamlines_cython)
    length_streamlines_cython = length(streamlines[::-1])
    length_streamlines_arrseq = length(Streamlines(streamlines)[::-1])
    assert_array_almost_equal(length_streamlines_arrseq, length_streamlines_cython)

    # Test streamlines having mixed dtype
    streamlines_mixed_dtype = [
        streamline,
        streamline.astype(np.float64),
        streamline.astype(np.int32),
        streamline.astype(np.int64),
    ]
    lengths_mixed_dtype = [length(s) for s in streamlines_mixed_dtype]
    assert_array_equal(length(streamlines_mixed_dtype), lengths_mixed_dtype)

    # Test streamlines with different shape
    length_streamlines_cython = length(heterogeneous_streamlines)

    for i, s in enumerate(heterogeneous_streamlines):
        length_streamline_python = length_python(s)
        assert_array_almost_equal(
            length_streamlines_cython[i], length_streamline_python
        )

    # Test streamline having integer dtype
    length_streamline = length(streamline.astype("int"))
    assert_equal(length_streamline.dtype, np.float64)

    # Test empty list
    assert_equal(length([]), 0.0)

    # Test streamline having only one point
    assert_equal(length(np.array([[1, 2, 3]])), 0.0)

    # We do not support list of lists, it should be numpy ndarray.
    streamline_unsupported = [[1, 2, 3], [4, 5, 5], [2, 1, 3], [4, 2, 1]]
    assert_raises(AttributeError, length, streamline_unsupported)

    # Test setting computing length of a numpy with flag WRITABLE=False
    streamlines_readonly = []
    for s in streamlines:
        streamlines_readonly.append(s.copy())
        streamlines_readonly[-1].setflags(write=False)

    assert_array_almost_equal(
        length(streamlines_readonly), [length_python(s) for s in streamlines_readonly]
    )
    streamlines_readonly = []
    for s in streamlines_64bit:
        streamlines_readonly.append(s.copy())
        streamlines_readonly[-1].setflags(write=False)

    assert_array_almost_equal(
        length(streamlines_readonly), [length_python(s) for s in streamlines_readonly]
    )


@set_random_number_generator(1234)
def test_length_memory_leaks(rng):
    # Test some dtypes
    dtypes = [np.float32, np.float64, np.int32, np.int64]
    for dtype in dtypes:
        s_rng = np.random.default_rng(1234)
        NB_STREAMLINES = 10000
        streamlines = [
            s_rng.standard_normal((s_rng.integers(10, 100), 3)).astype(dtype)
            for _ in range(NB_STREAMLINES)
        ]

        list_refcount_before = get_type_refcount()["list"]

        # lengths = length(streamlines)
        list_refcount_after = get_type_refcount()["list"]

        # Calling `length` shouldn't increase the refcount of `list`
        # since the return value is a numpy array.
        assert_equal(list_refcount_after, list_refcount_before)

    # Test mixed dtypes
    NB_STREAMLINES = 10000
    streamlines = []
    for i in range(NB_STREAMLINES):
        dtype = dtypes[i % len(dtypes)]
        streamlines.append(
            rng.standard_normal((rng.integers(10, 100), 3)).astype(dtype)
        )

    list_refcount_before = get_type_refcount()["list"]

    # lengths = length(streamlines)
    list_refcount_after = get_type_refcount()["list"]

    # Calling `length` shouldn't increase the refcount of `list`
    # since the return value is a numpy array.
    assert_equal(list_refcount_after, list_refcount_before)


@set_random_number_generator()
def test_unlist_relist_streamlines(rng):
    streamlines = [rng.random((10, 3)), rng.random((20, 3)), rng.random((5, 3))]
    points, offsets = unlist_streamlines(streamlines)
    assert_equal(offsets.dtype, np.dtype("i8"))
    assert_equal(points.shape, (35, 3))
    assert_equal(len(offsets), len(streamlines))

    streamlines2 = relist_streamlines(points, offsets)
    assert_equal(len(streamlines), len(streamlines2))
    for i in range(len(streamlines)):
        assert_array_equal(streamlines[i], streamlines2[i])


def test_transform_streamlines_dtype_in_place():
    identity = np.eye(4)
    streamlines = Streamlines([streamline])
    streamlines._data = streamlines._data.astype(np.float16)
    data_dtype = streamlines._data.dtype
    offsets_dtype = streamlines._offsets.dtype

    transform_streamlines(streamlines, identity, in_place=True)
    assert_equal(data_dtype, streamlines._data.dtype)
    assert_equal(offsets_dtype, streamlines._offsets.dtype)


def test_transform_streamlines_dtype():
    identity = np.eye(4)
    streamlines = Streamlines([streamline])
    streamlines._data = streamlines._data.astype(np.float16)
    data_dtype = streamlines._data.dtype
    offsets_dtype = streamlines._offsets.dtype

    streamlines = transform_streamlines(streamlines, identity, in_place=False)
    assert_equal(data_dtype, streamlines._data.dtype)
    assert_equal(offsets_dtype, streamlines._offsets.dtype)


def test_transform_empty_streamlines():
    identity = np.eye(4)
    streamlines = Streamlines([])

    streamlines = transform_streamlines(streamlines, identity, in_place=False)
    assert_equal(len(streamlines), 0)


@set_random_number_generator()
def test_deform_streamlines(rng):
    # Create Random deformation field
    deformation_field = rng.standard_normal((200, 200, 200, 3))
    stream2grid = np.array(
        [
            [-0.13152201, -0.52553149, -0.06759869, -0.80014208],
            [1.01579851, 0.19840874, 0.18875411, 0.81826065],
            [-0.07047617, -0.9290094, -0.55623385, 0.55165017],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )
    grid2world = np.array(
        [
            [0.83354727, 1.33876877, 1.0218087, 0.12809569],
            [0.83571344, 0.63824941, 0.20564267, 0.82740437],
            [-0.26574668, -0.66695577, 0.11636694, -0.02620037],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )
    stream2world = np.dot(stream2grid, grid2world)

    # Deform streamlines (let two grid spaces be the same for simplicity)
    new_streamlines = deform_streamlines(
        streamlines, deformation_field, stream2grid, grid2world, stream2grid, grid2world
    )

    # Interpolate displacements onto original streamlines
    streamlines_in_grid = transform_streamlines(streamlines, stream2grid)
    disps = values_from_volume(deformation_field, streamlines_in_grid, np.eye(4))

    # Put new_streamlines into world space
    new_streamlines_world = transform_streamlines(new_streamlines, stream2world)

    # Subtract disps from new_streamlines in world space
    orig_streamlines_world = np.subtract(
        np.array(new_streamlines_world, dtype=object), np.array(disps, dtype=object)
    )

    # Put orig_streamlines_world into voxmm
    orig_streamlines = transform_streamlines(
        orig_streamlines_world, np.linalg.inv(stream2world)
    )
    # All close because of floating pt imprecision
    for o, s in zip(orig_streamlines, streamlines):
        assert_allclose(s, o.astype(np.float32), rtol=1e-6, atol=1e-6)


def test_center_and_transform():
    A = np.array([[1, 2, 3], [1, 2, 3.0]])
    streamlines = [A for _ in range(10)]
    streamlines2, center = center_streamlines(streamlines)
    B = np.zeros((2, 3))
    assert_array_equal(streamlines2[0], B)
    assert_array_equal(center, A[0])

    affine = np.eye(4)
    affine[0, 0] = 2
    affine[:3, -1] = -np.array([2, 1, 1]) * center
    streamlines3 = transform_streamlines(streamlines, affine)
    assert_array_equal(streamlines3[0], B)


@set_random_number_generator()
def test_select_random_streamlines(rng):
    streamlines = [rng.random((10, 3)), rng.random((20, 3)), rng.random((5, 3))]
    new_streamlines = select_random_set_of_streamlines(streamlines, 2)
    assert_equal(len(new_streamlines), 2)

    new_streamlines = select_random_set_of_streamlines(streamlines, 4)
    assert_equal(len(new_streamlines), 3)


def compress_streamlines_python(streamline, tol_error=0.01, max_segment_length=10):
    """
    Python version of the FiberCompression found on
    https://github.com/scilus/FiberCompression.
    """
    if streamline.shape[0] <= 2:
        return streamline.copy()

    # Euclidean distance
    def segment_length(p1, p2):
        return np.sqrt(((p1 - p2) ** 2).sum())

    # Projection of a 3D point on a 3D line, minimal distance
    def dist_to_line(p1, p2, p0):
        return norm(np.cross(p2 - p1, p0 - p2)) / norm(p2 - p1)

    nb_points = 0
    compressed_streamline = np.zeros_like(streamline)

    # Copy first point since it is always kept.
    compressed_streamline[0, :] = streamline[0, :]
    nb_points += 1
    p1 = streamline[0]
    prev_id = 0

    for next_id, p2 in enumerate(streamline[2:], start=2):
        # Euclidean distance between last added point and current point.
        if segment_length(p1, p2) > max_segment_length:
            compressed_streamline[nb_points, :] = streamline[next_id - 1, :]
            nb_points += 1
            p1 = streamline[next_id - 1]
            prev_id = next_id - 1
            continue

        # Check that each point is not offset by more than `tol_error` mm.
        for _, p0 in enumerate(streamline[prev_id + 1 : next_id], start=prev_id + 1):
            dist = dist_to_line(p1, p2, p0)

            if np.isnan(dist) or dist > tol_error:
                compressed_streamline[nb_points, :] = streamline[next_id - 1, :]
                nb_points += 1
                p1 = streamline[next_id - 1]
                prev_id = next_id - 1
                break

    # Copy last point since it is always kept.
    compressed_streamline[nb_points, :] = streamline[-1, :]
    nb_points += 1

    # Make sure the array have the correct size
    return compressed_streamline[:nb_points]


def test_compress_streamlines():
    for compress_func in [compress_streamlines_python, compress_streamlines]:
        # Small streamlines (less than two points) are incompressible.
        for small_streamline in [
            np.array([[]]),
            np.array([[1, 1, 1]]),
            np.array([[1, 1, 1], [2, 2, 2]]),
        ]:
            c_streamline = compress_func(small_streamline)
            assert_equal(len(c_streamline), len(small_streamline))
            assert_array_equal(c_streamline, small_streamline)

        # Compressing a straight streamline that is less than 10mm long
        # should output a two points streamline.
        linear_streamline = np.linspace(0, 5, 100 * 3).reshape((100, 3))
        c_streamline = compress_func(linear_streamline)
        assert_equal(len(c_streamline), 2)
        assert_array_equal(c_streamline, [linear_streamline[0], linear_streamline[-1]])

        # The distance of consecutive points must be less or equal than some
        # value.
        max_segment_length = 10
        linear_streamline = np.linspace(0, 100, 100 * 3).reshape((100, 3))
        linear_streamline[:, 1:] = 0.0
        c_streamline = compress_func(
            linear_streamline, max_segment_length=max_segment_length
        )
        segments_length = np.sqrt((np.diff(c_streamline, axis=0) ** 2).sum(axis=1))
        assert_true(np.all(segments_length <= max_segment_length))
        assert_equal(len(c_streamline), 12)
        assert_array_equal(c_streamline, linear_streamline[::9])

        # A small `max_segment_length` should keep all points.
        c_streamline = compress_func(linear_streamline, max_segment_length=0.01)
        assert_array_equal(c_streamline, linear_streamline)

        # Test we can set `max_segment_length` to infinity
        # (like the C++ version)
        compress_func(streamline, max_segment_length=np.inf)

        # Incompressible streamline when `tol_error` == 1.
        simple_streamline = np.array(
            [[0, 0, 0], [1, 1, 0], [1.5, np.inf, 0], [2, 2, 0], [2.5, 20, 0], [3, 3, 0]]
        )

        # Because of np.inf, compressing that streamline causes a warning.
        with np.errstate(invalid="ignore"):
            c_streamline = compress_func(simple_streamline, tol_error=1)
            assert_array_equal(c_streamline, simple_streamline)

    # Create a special streamline where every other point is increasingly
    # farther from a straight line formed by the streamline endpoints.
    tol_errors = np.linspace(0, 10, 21)
    orthogonal_line = np.array([[-np.sqrt(2) / 2, np.sqrt(2) / 2, 0]], dtype=np.float32)
    special_streamline = np.array(
        [range(len(tol_errors) * 2 + 1)] * 3, dtype=np.float32
    ).T
    special_streamline[1::2] += orthogonal_line * tol_errors[:, None]

    # # Uncomment to see the streamline.
    # import pylab as plt
    # plt.plot(special_streamline[:, 0], special_streamline[:, 1], '.-')
    # plt.axis('equal'); plt.show()

    # Test different values for `tol_error`.
    for i, tol_error in enumerate(tol_errors):
        cspecial_streamline = compress_streamlines(
            special_streamline, tol_error=tol_error + 1e-4, max_segment_length=np.inf
        )

        # First and last points should always be the same as the original ones.
        assert_array_equal(cspecial_streamline[0], special_streamline[0])
        assert_array_equal(cspecial_streamline[-1], special_streamline[-1])

        assert_equal(len(cspecial_streamline), len(special_streamline) - ((i * 2) + 1))

        # Make sure Cython and Python versions are the same.
        cstreamline_python = compress_streamlines_python(
            special_streamline, tol_error=tol_error + 1e-4, max_segment_length=np.inf
        )
        assert_equal(len(cspecial_streamline), len(cstreamline_python))
        assert_array_almost_equal(cspecial_streamline, cstreamline_python)


def test_compress_streamlines_identical_points():
    sl_1 = np.array([[1, 1, 1], [1, 1, 1], [2, 2, 2], [3, 3, 3], [3, 3, 3]])
    sl_2 = np.array([[1, 1, 1], [1, 1, 1], [1, 1, 1], [2, 2, 2]])
    sl_3 = np.array(
        [
            [1, 1, 1],
            [1, 1, 1],
            [1, 1, 1],
            [1, 1, 1],
            [2, 2, 2],
            [2, 2, 2],
            [2, 2, 2],
            [3, 3, 3],
            [3, 3, 3],
        ]
    )
    sl_4 = np.array(
        [
            [1, 1, 1],
            [1, 1, 1],
            [1, 1, 1],
            [1, 1, 1],
            [2, 2, 2],
            [3, 3, 3],
            [3, 3, 3],
            [1, 1, 1],
        ]
    )
    new_sl_1 = compress_streamlines(sl_1)
    new_sl_2 = compress_streamlines(sl_2)
    new_sl_3 = compress_streamlines(sl_3)
    new_sl_4 = compress_streamlines(sl_4)
    npt.assert_array_equal(new_sl_1, np.array([[1, 1, 1], [3, 3, 3]]))
    npt.assert_array_equal(new_sl_2, np.array([[1, 1, 1], [2, 2, 2]]))
    npt.assert_array_equal(new_sl_3, new_sl_1)
    npt.assert_array_equal(new_sl_4, np.array([[1, 1, 1], [3, 3, 3], [1, 1, 1]]))


@set_random_number_generator(1234)
def test_compress_streamlines_memory_leaks(rng):
    # Test some dtypes
    dtypes = [np.float32, np.float64, np.int32, np.int64]
    for dtype in dtypes:
        s_rng = np.random.default_rng(1234)
        NB_STREAMLINES = 10000
        streamlines = [
            s_rng.standard_normal((s_rng.integers(10, 100), 3)).astype(dtype)
            for _ in range(NB_STREAMLINES)
        ]

        list_refcount_before = get_type_refcount()["list"]

        cstreamlines = compress_streamlines(streamlines)
        list_refcount_after = get_type_refcount()["list"]
        del (
            cstreamlines
        )  # Delete `cstreamlines` because it holds a reference to `list`   # noqa: F841

        # Calling `compress_streamlines` should increase the refcount of `list`
        # by one since we kept the returned value.
        assert_equal(list_refcount_after, list_refcount_before + 1)

    # Test mixed dtypes
    NB_STREAMLINES = 10000
    streamlines = []
    for i in range(NB_STREAMLINES):
        dtype = dtypes[i % len(dtypes)]
        streamlines.append(
            rng.standard_normal((rng.integers(10, 100), 3)).astype(dtype)
        )

    list_refcount_before = get_type_refcount()["list"]
    cstreamlines = compress_streamlines(streamlines)  # noqa: F841
    list_refcount_after = get_type_refcount()["list"]

    # Calling `compress_streamlines` should increase the refcount of `list` by
    # one since we kept the returned value.
    assert_equal(list_refcount_after, list_refcount_before + 1)


def generate_sl(streamlines):
    """
    Helper function that takes a sequence and returns a generator

    Parameters
    ----------
    streamlines : sequence
        Usually, this would be a list of 2D arrays, representing streamlines

    Returns
    -------
    generator
    """
    for sl in streamlines:
        yield sl


def test_select_by_rois():
    streamlines = [
        np.array([[0, 0.0, 0.9], [1.9, 0.0, 0.0]]),
        np.array([[0.1, 0.0, 0], [0, 1.0, 1.0], [0, 2.0, 2.0]]),
        np.array([[2, 2, 2], [3, 3, 3]]),
    ]

    # Make two ROIs:
    mask1 = np.zeros((4, 4, 4), dtype=bool)
    mask2 = np.zeros_like(mask1)
    mask1[0, 0, 0] = True
    mask2[1, 0, 0] = True

    selection = select_by_rois(streamlines, np.eye(4), [mask1], [True], tol=1)

    assert_arrays_equal(list(selection), [streamlines[0], streamlines[1]])

    selection = select_by_rois(
        streamlines, np.eye(4), [mask1, mask2], [True, True], tol=1
    )

    assert_arrays_equal(list(selection), [streamlines[0], streamlines[1]])

    selection = select_by_rois(streamlines, np.eye(4), [mask1, mask2], [True, False])

    assert_arrays_equal(list(selection), [streamlines[1]])

    # Setting tolerance too low gets overridden:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        selection = select_by_rois(
            streamlines, np.eye(4), [mask1, mask2], [True, False], tol=0.1
        )

        assert_arrays_equal(list(selection), [streamlines[1]])

    selection = select_by_rois(
        streamlines, np.eye(4), [mask1, mask2], [True, True], tol=0.87
    )

    assert_arrays_equal(list(selection), [streamlines[1]])

    mask3 = np.zeros_like(mask1)
    mask3[0, 2, 2] = 1
    selection = select_by_rois(
        streamlines, np.eye(4), [mask1, mask2, mask3], [True, True, False], tol=1.0
    )

    assert_arrays_equal(list(selection), [streamlines[0]])

    # Select using only one ROI
    selection = select_by_rois(streamlines, np.eye(4), [mask1], [True], tol=0.87)
    assert_arrays_equal(list(selection), [streamlines[1]])

    selection = select_by_rois(streamlines, np.eye(4), [mask1], [True], tol=1.0)
    assert_arrays_equal(list(selection), [streamlines[0], streamlines[1]])

    # Use different modes:
    selection = select_by_rois(
        streamlines,
        np.eye(4),
        [mask1, mask2, mask3],
        [True, True, False],
        mode="all",
        tol=1.0,
    )
    assert_arrays_equal(list(selection), [streamlines[0]])

    selection = select_by_rois(
        streamlines,
        np.eye(4),
        [mask1, mask2, mask3],
        [True, True, False],
        mode="either_end",
        tol=1.0,
    )
    assert_arrays_equal(list(selection), [streamlines[0]])

    selection = select_by_rois(
        streamlines,
        np.eye(4),
        [mask1, mask2, mask3],
        [True, True, False],
        mode="both_end",
        tol=1.0,
    )
    assert_arrays_equal(list(selection), [streamlines[0]])

    mask2[0, 2, 2] = True
    selection = select_by_rois(
        streamlines,
        np.eye(4),
        [mask1, mask2, mask3],
        [True, True, False],
        mode="both_end",
        tol=1.0,
    )

    assert_arrays_equal(list(selection), [streamlines[0], streamlines[1]])

    # Test with generator input:
    selection = select_by_rois(
        generate_sl(streamlines), np.eye(4), [mask1], [True], tol=1.0
    )
    assert_arrays_equal(list(selection), [streamlines[0], streamlines[1]])


def test_orient_by_rois():
    streamlines = Streamlines(
        [
            np.array([[0, 0.0, 0], [1, 0.0, 0.0], [2, 0.0, 0.0]]),
            np.array([[2, 0.0, 0.0], [1, 0.0, 0], [0, 0, 0.0]]),
        ]
    )

    # Make two ROIs:
    mask1_vol = np.zeros((4, 4, 4), dtype=bool)
    mask2_vol = np.zeros_like(mask1_vol)
    mask1_vol[0, 0, 0] = True
    mask2_vol[1, 0, 0] = True
    mask1_coords = np.array(np.where(mask1_vol)).T
    mask2_coords = np.array(np.where(mask2_vol)).T

    # If there is an affine, we'll use it:
    affine = np.eye(4)
    affine[:, 3] = [-1, 100, -20, 1]
    # Transform the streamlines:
    x_streamlines = Streamlines([sl + affine[:3, 3] for sl in streamlines])

    # After reorientation, this should be the answer:
    flipped_sl = Streamlines([streamlines[0], streamlines[1][::-1]])
    new_streamlines = orient_by_rois(
        streamlines, np.eye(4), mask1_vol, mask2_vol, in_place=False, as_generator=False
    )
    npt.assert_array_equal(new_streamlines, flipped_sl)

    npt.assert_(new_streamlines is not streamlines)

    # Test with affine:
    x_flipped_sl = Streamlines([s + affine[:3, 3] for s in flipped_sl])
    new_streamlines = orient_by_rois(
        x_streamlines, affine, mask1_vol, mask2_vol, in_place=False, as_generator=False
    )
    npt.assert_array_equal(new_streamlines, x_flipped_sl)
    npt.assert_(new_streamlines is not x_streamlines)

    # Test providing coord ROIs instead of vol ROIs:
    new_streamlines = orient_by_rois(
        x_streamlines,
        affine,
        mask1_coords,
        mask2_coords,
        in_place=False,
        as_generator=False,
    )
    npt.assert_array_equal(new_streamlines, x_flipped_sl)

    # Test with as_generator set to True
    new_streamlines = orient_by_rois(
        streamlines, np.eye(4), mask1_vol, mask2_vol, in_place=False, as_generator=True
    )

    npt.assert_(isinstance(new_streamlines, types.GeneratorType))
    ll = Streamlines(new_streamlines)
    npt.assert_array_equal(ll, flipped_sl)

    # Test with as_generator set to True and with the affine
    new_streamlines = orient_by_rois(
        x_streamlines, affine, mask1_vol, mask2_vol, in_place=False, as_generator=True
    )

    npt.assert_(isinstance(new_streamlines, types.GeneratorType))
    ll = Streamlines(new_streamlines)
    npt.assert_array_equal(ll, x_flipped_sl)

    # Test with generator input:
    new_streamlines = orient_by_rois(
        generate_sl(streamlines),
        np.eye(4),
        mask1_vol,
        mask2_vol,
        in_place=False,
        as_generator=True,
    )

    npt.assert_(isinstance(new_streamlines, types.GeneratorType))
    ll = Streamlines(new_streamlines)
    npt.assert_array_equal(ll, flipped_sl)

    # Generator output cannot take a True `in_place` kwarg:
    npt.assert_raises(
        ValueError,
        orient_by_rois,
        *[generate_sl(streamlines), np.eye(4), mask1_vol, mask2_vol],
        **{"in_place": True, "as_generator": True},
    )

    # But you can input a generator and get a non-generator as output:
    new_streamlines = orient_by_rois(
        generate_sl(streamlines),
        np.eye(4),
        mask1_vol,
        mask2_vol,
        in_place=False,
        as_generator=False,
    )

    npt.assert_(not isinstance(new_streamlines, types.GeneratorType))
    npt.assert_array_equal(new_streamlines, flipped_sl)

    # Modify in-place:
    new_streamlines = orient_by_rois(
        streamlines, np.eye(4), mask1_vol, mask2_vol, in_place=True, as_generator=False
    )

    npt.assert_array_equal(new_streamlines, flipped_sl)
    # The two objects are one and the same:
    npt.assert_(new_streamlines is streamlines)


def test_orient_by_streamline():
    streamlines = Streamlines(
        [
            np.array([[0, 0.0, 0], [1, 0.0, 0.0], [2, 0.0, 0.0]]),
            np.array([[2, 0.0, 0.0], [1, 0.0, 0], [0, 0, 0.0]]),
        ]
    )

    # If there is an affine, we'll use it:
    affine = np.eye(4)
    affine[:, 3] = [-1, 100, -20, 1]
    # Transform the streamlines:
    x_streamlines = Streamlines([sl + affine[:3, 3] for sl in streamlines])

    standard_streamline = streamlines[0]

    # After reorientation, this should be the answer:
    flipped_sl = Streamlines([streamlines[0], streamlines[1][::-1]])

    new_streamlines = orient_by_streamline(
        streamlines, standard_streamline, n_points=12, in_place=False
    )

    npt.assert_array_equal(new_streamlines, flipped_sl)
    npt.assert_(new_streamlines is not streamlines)

    # Test with affine:
    x_flipped_sl = Streamlines([s + affine[:3, 3] for s in flipped_sl])
    new_streamlines = orient_by_streamline(
        x_streamlines, standard_streamline, in_place=False
    )
    npt.assert_array_equal(new_streamlines, x_flipped_sl)
    npt.assert_(new_streamlines is not x_streamlines)

    # Test with as_generator set to True
    new_streamlines = orient_by_streamline(
        streamlines, standard_streamline, in_place=False, as_generator=True
    )

    npt.assert_(isinstance(new_streamlines, types.GeneratorType))
    ll = Streamlines(new_streamlines)
    npt.assert_array_equal(ll, flipped_sl)

    # Test with as_generator set to True and with the affine
    new_streamlines = orient_by_streamline(
        x_streamlines, standard_streamline, in_place=False, as_generator=True
    )

    npt.assert_(isinstance(new_streamlines, types.GeneratorType))
    ll = Streamlines(new_streamlines)
    npt.assert_array_equal(ll, x_flipped_sl)

    # Modify in-place:
    new_streamlines = orient_by_streamline(
        streamlines, standard_streamline, in_place=True
    )

    npt.assert_array_equal(new_streamlines, flipped_sl)
    # The two objects are one and the same:
    npt.assert_(new_streamlines is streamlines)


def test_values_from_volume():
    decimal = 4
    data3d = np.arange(2000).reshape(20, 10, 10)
    # Test two cases of 4D data (handled differently)
    # One where the last dimension is length 3:
    data4d_3vec = np.arange(6000).reshape(20, 10, 10, 3)
    # The other where the last dimension is not 3:
    data4d_2vec = np.arange(4000).reshape(20, 10, 10, 2)
    for dt in [np.float32, np.float64]:
        for data in [data3d, data4d_3vec, data4d_2vec]:
            sl1 = [
                np.array([[1, 0, 0], [1.5, 0, 0], [2, 0, 0], [2.5, 0, 0]]).astype(dt),
                np.array([[2, 0, 0], [3.1, 0, 0], [3.9, 0, 0], [4.1, 0, 0]]).astype(dt),
            ]

            ans1 = [
                [
                    data[1, 0, 0],
                    data[1, 0, 0] + (data[2, 0, 0] - data[1, 0, 0]) / 2,
                    data[2, 0, 0],
                    data[2, 0, 0] + (data[3, 0, 0] - data[2, 0, 0]) / 2,
                ],
                [
                    data[2, 0, 0],
                    data[3, 0, 0] + (data[4, 0, 0] - data[3, 0, 0]) * 0.1,
                    data[3, 0, 0] + (data[4, 0, 0] - data[3, 0, 0]) * 0.9,
                    data[4, 0, 0] + (data[5, 0, 0] - data[4, 0, 0]) * 0.1,
                ],
            ]

            vv = values_from_volume(data, sl1, np.eye(4))
            npt.assert_almost_equal(vv, ans1, decimal=decimal)

            vv = values_from_volume(data, np.array(sl1), np.eye(4))
            npt.assert_almost_equal(vv, ans1, decimal=decimal)

            vv = values_from_volume(data, Streamlines(sl1), np.eye(4))
            npt.assert_almost_equal(vv, ans1, decimal=decimal)

            affine = np.eye(4)
            affine[:, 3] = [-100, 10, 1, 1]
            x_sl1 = transform_streamlines(sl1, affine)
            x_sl2 = transform_streamlines(sl1, affine)

            vv = values_from_volume(data, x_sl1, affine)
            npt.assert_almost_equal(vv, ans1, decimal=decimal)

            x_sl1 = transform_streamlines(sl1, affine)
            vv = values_from_volume(data, x_sl1, affine)

            npt.assert_almost_equal(vv, ans1, decimal=decimal)

            # Test that the streamlines haven't mutated:
            l_sl2 = list(x_sl2)
            npt.assert_equal(x_sl1, l_sl2)

            vv = values_from_volume(data, np.array(x_sl1), affine)
            npt.assert_almost_equal(vv, ans1, decimal=decimal)
            npt.assert_equal(np.array(x_sl1), np.array(l_sl2))

            # Test for lists of streamlines with different numbers of nodes:
            sl2 = [sl1[0][:-1], sl1[1]]
            ans2 = [ans1[0][:-1], ans1[1]]
            vv = values_from_volume(data, sl2, np.eye(4))
            for ii, v in enumerate(vv):
                npt.assert_almost_equal(v, ans2[ii], decimal=decimal)

    # We raise an error if the streamlines fed don't make sense. In this
    # case, a tuple instead of a list, generator or array
    nonsense_sl = (
        np.array([[1, 0, 0], [1.5, 0, 0], [2, 0, 0], [2.5, 0, 0]]),
        np.array([[2, 0, 0], [3.1, 0, 0], [3.9, 0, 0], [4.1, 0, 0]]),
    )

    npt.assert_raises(RuntimeError, values_from_volume, data, nonsense_sl, np.eye(4))

    # For some use-cases we might have singleton streamlines (with only one
    # node each):
    data3D = np.ones((2, 2, 2))
    streamlines = np.ones((10, 1, 3))
    npt.assert_equal(values_from_volume(data3D, streamlines, np.eye(4)).shape, (10, 1))
    data4D = np.ones((2, 2, 2, 2))
    streamlines = np.ones((10, 1, 3))
    npt.assert_equal(
        values_from_volume(data4D, streamlines, np.eye(4)).shape, (10, 1, 2)
    )


def test_streamlines_generator():
    # Test generator
    streamlines_generator = Streamlines(generate_sl(streamlines))
    npt.assert_equal(len(streamlines_generator), len(streamlines))
    # Nothing should change
    streamlines_generator.append(np.array([]))
    npt.assert_equal(len(streamlines_generator), len(streamlines))

    # Test append error
    npt.assert_raises(
        ValueError, streamlines_generator.append, np.array(streamlines, dtype=object)
    )

    # Test empty streamlines
    streamlines_generator = Streamlines(np.array([]))
    npt.assert_equal(len(streamlines_generator), 0)


def test_cluster_confidence():
    mysl = np.array([np.arange(10)] * 3, "float").T

    # a short streamline (<20 mm) should raise an error unless override=True
    test_streamlines = Streamlines()
    test_streamlines.append(mysl)
    assert_raises(ValueError, cluster_confidence, test_streamlines)
    cci = cluster_confidence(test_streamlines, override=True)

    # two identical streamlines should raise an error
    test_streamlines = Streamlines()
    test_streamlines.append(mysl, cache_build=True)
    test_streamlines.append(mysl)
    test_streamlines.finalize_append()
    assert_raises(ValueError, cluster_confidence, test_streamlines)

    # 3 offset collinear streamlines
    test_streamlines = Streamlines()
    test_streamlines.append(mysl, cache_build=True)
    test_streamlines.append(mysl + 1)
    test_streamlines.append(mysl + 2)
    test_streamlines.finalize_append()

    cci = cluster_confidence(test_streamlines, override=True)

    assert_almost_equal(cci[0], cci[2])
    assert_true(cci[1] > cci[0])

    # 3 parallel streamlines
    mysl = np.zeros([10, 3])
    mysl[:, 0] = np.arange(10)
    mysl2 = mysl.copy()
    mysl2[:, 1] = 1
    mysl3 = mysl.copy()
    mysl3[:, 1] = 2
    mysl4 = mysl.copy()
    mysl4[:, 1] = 4
    mysl5 = mysl.copy()
    mysl5[:, 1] = 5000

    test_streamlines_p1 = Streamlines()
    test_streamlines_p1.append(mysl, cache_build=True)
    test_streamlines_p1.append(mysl2)
    test_streamlines_p1.append(mysl3)
    test_streamlines_p1.finalize_append()
    test_streamlines_p2 = Streamlines()
    test_streamlines_p2.append(mysl, cache_build=True)
    test_streamlines_p2.append(mysl3)
    test_streamlines_p2.append(mysl4)
    test_streamlines_p2.finalize_append()
    test_streamlines_p3 = Streamlines()
    test_streamlines_p3.append(mysl, cache_build=True)
    test_streamlines_p3.append(mysl2)
    test_streamlines_p3.append(mysl3)
    test_streamlines_p3.append(mysl5)
    test_streamlines_p3.finalize_append()

    cci_p1 = cluster_confidence(test_streamlines_p1, override=True)
    cci_p2 = cluster_confidence(test_streamlines_p2, override=True)

    # test relative distance
    assert_array_equal(cci_p1, cci_p2 * 2)

    # test simple cci calculation
    expected_p1 = np.array([1.0 / 1 + 1.0 / 2, 1.0 / 1 + 1.0 / 1, 1.0 / 1 + 1.0 / 2])
    expected_p2 = np.array([1.0 / 2 + 1.0 / 4, 1.0 / 2 + 1.0 / 2, 1.0 / 2 + 1.0 / 4])
    assert_array_equal(expected_p1, cci_p1)
    assert_array_equal(expected_p2, cci_p2)

    # test power variable calculation (dropoff with distance)
    cci_p1_pow2 = cluster_confidence(test_streamlines_p1, power=2, override=True)

    expected_p1_pow2 = np.array(
        [
            np.power(1.0 / 1, 2) + np.power(1.0 / 2, 2),
            np.power(1.0 / 1, 2) + np.power(1.0 / 1, 2),
            np.power(1.0 / 1, 2) + np.power(1.0 / 2, 2),
        ]
    )

    assert_array_equal(cci_p1_pow2, expected_p1_pow2)

    # test max distance (ignore distant sls)
    cci_dist = cluster_confidence(test_streamlines_p3, max_mdf=5, override=True)

    expected_cci_dist = np.concatenate([cci_p1, np.zeros(1)])
    assert_array_equal(cci_dist, expected_cci_dist)



"""
===============================
Particle Filtering Tractography
===============================
Particle Filtering Tractography (PFT) :footcite:p:`Girard2014` uses tissue
partial volume estimation (PVE) to reconstruct trajectories connecting the gray
matter, and not incorrectly stopping in the white matter or in the corticospinal
fluid. It relies on a stopping criterion that identifies the tissue where the
streamline stopped. If the streamline correctly stopped in the gray matter, the
trajectory is kept. If the streamline incorrectly stopped in the white matter
or in the corticospinal fluid, PFT uses anatomical information to find an
alternative streamline segment to extend the trajectory. When this segment is
found, the tractography continues until the streamline correctly stops in the
gray matter.

PFT finds an alternative streamline segment whenever the stopping criterion
returns a position classified as 'INVALIDPOINT'.

This example is an extension of
:ref:`sphx_glr_examples_built_fiber_tracking_tracking_probabilistic.py` and
:ref:`sphx_glr_examples_built_fiber_tracking_tracking_stopping_criterion.py`
examples. We begin by loading the data, fitting a Constrained Spherical
Deconvolution (CSD) reconstruction model, creating the probabilistic direction
getter and defining the seeds.
"""

import numpy as np

from dipy.core.gradients import gradient_table
from dipy.data import default_sphere, get_fnames
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, load_nifti_data
from dipy.io.stateful_tractogram import Space, StatefulTractogram
from dipy.io.streamline import save_trk
from dipy.reconst.csdeconv import ConstrainedSphericalDeconvModel, auto_response_ssst
from dipy.tracking import utils
from dipy.tracking.stopping_criterion import CmcStoppingCriterion
from dipy.tracking.streamline import Streamlines
from dipy.tracking.tracker import pft_tracking, probabilistic_tracking
from dipy.viz import actor, colormap, has_fury, window

# Enables/disables interactive visualization
interactive = False

hardi_fname, hardi_bval_fname, hardi_bvec_fname = get_fnames(name="stanford_hardi")
label_fname = get_fnames(name="stanford_labels")
f_pve_csf, f_pve_gm, f_pve_wm = get_fnames(name="stanford_pve_maps")

data, affine, hardi_img = load_nifti(hardi_fname, return_img=True)
labels = load_nifti_data(label_fname)
bvals, bvecs = read_bvals_bvecs(hardi_bval_fname, hardi_bvec_fname)
gtab = gradient_table(bvals, bvecs=bvecs)

pve_csf_data = load_nifti_data(f_pve_csf)
pve_gm_data = load_nifti_data(f_pve_gm)
pve_wm_data, _, voxel_size = load_nifti(f_pve_wm, return_voxsize=True)

shape = labels.shape

response, ratio = auto_response_ssst(gtab, data, roi_radii=10, fa_thr=0.7)
csd_model = ConstrainedSphericalDeconvModel(gtab, response)
csd_fit = csd_model.fit(data, mask=pve_wm_data)

seed_mask = labels == 2
seed_mask[pve_wm_data < 0.5] = 0
seeds = utils.seeds_from_mask(seed_mask, affine, density=2)

###############################################################################
# CMC/ACT Stopping Criterion
# ==========================
# Continuous map criterion (CMC) :footcite:p:`Girard2014` and
# Anatomically-constrained tractography (ACT) :footcite:p:`Smith2012` both use
# PVEs information from anatomical images to determine when the tractography
# stops. Both stopping criterion use a trilinear interpolation at the tracking
# position. CMC stopping criterion uses a probability derived from the PVE maps
# to determine if the streamline reaches a 'valid' or 'invalid' region. ACT uses
# a fixed threshold on the PVE maps. Both stopping criterion can be used in
# conjunction with PFT. In this example, we used CMC.

voxel_size = np.average(voxel_size[1:4])
step_size = 0.2

cmc_criterion = CmcStoppingCriterion.from_pve(
    pve_wm_data,
    pve_gm_data,
    pve_csf_data,
    step_size=step_size,
    average_voxel_size=voxel_size,
)

###############################################################################
# Particle Filtering Tractography
# ===============================
# `pft_back_tracking_dist` is the distance in mm to backtrack when the
# tractography incorrectly stops in the WM or CSF. `pft_front_tracking_dist`
# is the distance in mm to track after the stopping event using PFT.
#
# The `particle_count` parameter is the number of samples used in the particle
# filtering algorithm.
#
# `min_wm_pve_before_stopping` controls when the tracking can stop in the GM.
# The tractography must reaches a position where WM_pve >=
# `min_wm_pve_before_stopping` before stopping in the GM. As long as the
# condition is not reached and WM_pve > 0, the tractography will continue.
# It is particularlyusefull to set this parameter > 0.5 when seeding
# tractography at the WM-GM interface or in the sub-cortical GM. It allows
# streamlines to exit the seeding voxels until they reach the deep white
# matter where WM_pve > `min_wm_pve_before_stopping`.
pft_streamline_gen = pft_tracking(
    seeds,
    cmc_criterion,
    affine,
    max_cross=1,
    step_size=step_size,
    max_len=1000,
    pft_back_tracking_dist=2,
    pft_front_tracking_dist=1,
    particle_count=15,
    return_all=False,
    min_wm_pve_before_stopping=1,
    max_angle=20.0,
    sphere=default_sphere,
    sh=csd_fit.shm_coeff,
)
streamlines = Streamlines(pft_streamline_gen)
sft = StatefulTractogram(streamlines, hardi_img, Space.RASMM)
save_trk(sft, "tractogram_pft.trk")

if has_fury:
    scene = window.Scene()
    scene.add(actor.line(streamlines, colors=colormap.line_colors(streamlines)))
    window.record(scene=scene, out_path="tractogram_pft.png", size=(800, 800))
    if interactive:
        window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Corpus Callosum using particle filtering tractography

# Local Probabilistic Tractography
prob_streamline_generator = probabilistic_tracking(
    seeds,
    cmc_criterion,
    affine,
    step_size=step_size,
    max_len=1000,
    return_all=False,
    sh=csd_fit.shm_coeff,
    max_angle=20.0,
    sphere=default_sphere,
)
streamlines = Streamlines(prob_streamline_generator)
sft = StatefulTractogram(streamlines, hardi_img, Space.RASMM)
save_trk(sft, "tractogram_probabilistic_cmc.trk")

if has_fury:
    scene = window.Scene()
    scene.add(actor.line(streamlines, colors=colormap.line_colors(streamlines)))
    window.record(
        scene=scene, out_path="tractogram_probabilistic_cmc.png", size=(800, 800)
    )
    if interactive:
        window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Corpus Callosum using probabilistic tractography
#
#
#
# References
# ----------
#
# .. footbibliography::
#



import numpy as np
from numpy.testing import (
    assert_,
    assert_almost_equal,
    assert_array_almost_equal,
    assert_equal,
    assert_warns,
)

import dipy.core.gradients as dpg
import dipy.data as dpd
from dipy.denoise.noise_estimate import (
    _inv_nchi_cdf,
    _piesno_3D,
    estimate_sigma,
    piesno,
)
from dipy.denoise.pca_noise_estimate import pca_noise_estimate
from dipy.io.image import load_nifti_data
from dipy.testing.decorators import set_random_number_generator


def test_inv_nchi():
    # See page 5 of the reference paper for tested values
    # Values taken from hispeed.MedianPIESNO.lambdaPlus
    # and hispeed.MedianPIESNO.lambdaMinus
    N = 8
    K = 20
    alpha = 0.01

    lambdaMinus = _inv_nchi_cdf(N, K, alpha / 2)
    lambdaPlus = _inv_nchi_cdf(N, K, 1 - alpha / 2)

    assert_almost_equal(lambdaMinus, 6.464855180579397)
    assert_almost_equal(lambdaPlus, 9.722849086419043)


@set_random_number_generator()
def test_piesno(rng):
    # Values taken from hispeed.OptimalPIESNO with the test data
    # in the package computed in matlab
    test_piesno_data = load_nifti_data(dpd.get_fnames(name="test_piesno"))
    sigma = piesno(
        test_piesno_data, N=8, alpha=0.01, step=1, eps=1e-10, return_mask=False
    )
    assert_almost_equal(sigma, 0.010749458025559)

    noise1 = (rng.standard_normal((100, 100, 100)) * 50) + 10
    noise2 = (rng.standard_normal((100, 100, 100)) * 50) + 10
    rician_noise = np.sqrt(noise1**2 + noise2**2)
    sigma, mask = piesno(
        rician_noise, N=1, alpha=0.01, step=1, eps=1e-10, return_mask=True
    )

    # less than 3% of error?
    assert_(np.abs(sigma - 50) / sigma < 0.03)

    # Test using the median as the initial estimation
    initial_estimation = np.median(sigma) / np.sqrt(2 * _inv_nchi_cdf(1, 1, 0.5))

    sigma, mask = _piesno_3D(
        rician_noise,
        N=1,
        alpha=0.01,
        step=1,
        eps=1e-10,
        return_mask=True,
        initial_estimation=initial_estimation,
    )

    assert_(np.abs(sigma - 50) / sigma < 0.03)

    sigma = _piesno_3D(
        rician_noise,
        N=1,
        alpha=0.01,
        step=1,
        eps=1e-10,
        return_mask=False,
        initial_estimation=initial_estimation,
    )
    assert_(np.abs(sigma - 50) / sigma < 0.03)

    sigma = _piesno_3D(
        np.zeros_like(rician_noise),
        N=1,
        alpha=0.01,
        step=1,
        eps=1e-10,
        return_mask=False,
        initial_estimation=initial_estimation,
    )

    assert_(np.all(sigma == 0))

    sigma, mask = _piesno_3D(
        np.zeros_like(rician_noise),
        N=1,
        alpha=0.01,
        step=1,
        eps=1e-10,
        return_mask=True,
        initial_estimation=initial_estimation,
    )

    assert_(np.all(sigma == 0))
    assert_(np.all(mask == 0))

    # Check if no noise points found in array it exits
    sigma = _piesno_3D(
        1000 * np.ones_like(rician_noise),
        N=1,
        alpha=0.01,
        step=1,
        eps=1e-10,
        return_mask=False,
        initial_estimation=10,
    )
    assert_(np.all(sigma == 10))


def test_piesno_type():
    # This is testing if the `sum_m2` cast is overflowing
    data = np.ones((10, 10, 10), dtype=np.int16)
    for i in range(10):
        data[:, i, :] = i * 26

    sigma = piesno(data, N=2, alpha=0.01, step=1, eps=1e-10, return_mask=False)
    assert_almost_equal(sigma, 79.970003117424739)


def test_estimate_sigma():
    sigma = estimate_sigma(np.ones((7, 7, 7)), disable_background_masking=True)
    assert_equal(sigma, 0.0)

    sigma = estimate_sigma(np.ones((7, 7, 7, 3)), disable_background_masking=True)
    assert_equal(sigma, np.array([0.0, 0.0, 0.0]))

    sigma = estimate_sigma(5 * np.ones((7, 7, 7)), disable_background_masking=False)
    assert_equal(sigma, 0.0)

    sigma = estimate_sigma(5 * np.ones((7, 7, 7, 3)), disable_background_masking=False)
    assert_equal(sigma, np.array([0.0, 0.0, 0.0]))

    arr = np.zeros((3, 3, 3))
    arr[0, 0, 0] = 1
    sigma = estimate_sigma(arr, disable_background_masking=False, N=1)
    assert_array_almost_equal(
        sigma, (0.10286889997472792 / np.sqrt(0.42920367320510366))
    )

    arr = np.zeros((3, 3, 3, 3))
    arr[0, 0, 0] = 1
    sigma = estimate_sigma(arr, disable_background_masking=False, N=1)
    assert_array_almost_equal(
        sigma,
        np.array(
            [
                0.10286889997472792 / np.sqrt(0.42920367320510366),
                0.10286889997472792 / np.sqrt(0.42920367320510366),
                0.10286889997472792 / np.sqrt(0.42920367320510366),
            ]
        ),
    )

    arr = np.zeros((3, 3, 3))
    arr[0, 0, 0] = 1
    sigma = estimate_sigma(arr, disable_background_masking=True, N=4)
    assert_array_almost_equal(sigma, 0.46291005 / np.sqrt(0.4834941393603609))

    arr = np.zeros((3, 3, 3))
    arr[0, 0, 0] = 1
    sigma = estimate_sigma(arr, disable_background_masking=True, N=0)
    assert_array_almost_equal(sigma, 0.46291005 / np.sqrt(1))
    arr = np.zeros((3, 3, 3, 3))

    arr[0, 0, 0] = 1
    sigma = estimate_sigma(arr, disable_background_masking=True, N=12)
    assert_array_almost_equal(
        sigma,
        np.array(
            [
                0.46291005 / np.sqrt(0.4946862482541263),
                0.46291005 / np.sqrt(0.4946862482541263),
                0.46291005 / np.sqrt(0.4946862482541263),
            ]
        ),
    )


@set_random_number_generator(1984)
def test_pca_noise_estimate(rng):
    # MUBE:
    bvals1 = np.concatenate([np.zeros(17), np.ones(3) * 1000])
    bvecs1 = np.concatenate([np.zeros((17, 3)), np.eye(3)])
    gtab1 = dpg.gradient_table(bvals1, bvecs=bvecs1)
    # SIBE:
    bvals2 = np.concatenate([np.zeros(1), np.ones(3) * 1000])
    bvecs2 = np.concatenate([np.zeros((1, 3)), np.eye(3)])
    gtab2 = dpg.gradient_table(bvals2, bvecs=bvecs2)

    for images_as_samples in [True, False]:
        for patch_radius in [1, 2]:
            for gtab in [gtab1, gtab2]:
                for dtype in [np.int16, np.float64]:
                    signal = np.ones((20, 20, 20, gtab.bvals.shape[0]))
                    for correct_bias in [True, False]:
                        if not correct_bias:
                            # High signal for no bias correction
                            signal = signal * 100

                        sigma = 1
                        noise1 = rng.normal(0, sigma, size=signal.shape)
                        noise2 = rng.normal(0, sigma, size=signal.shape)

                        # Rician noise:
                        data = np.sqrt((signal + noise1) ** 2 + noise2**2)

                        sigma_est = pca_noise_estimate(
                            data.astype(dtype),
                            gtab,
                            correct_bias=correct_bias,
                            patch_radius=patch_radius,
                            images_as_samples=images_as_samples,
                        )
                        # print("sigma_est:", sigma_est)
                        assert_array_almost_equal(np.mean(sigma_est), sigma, decimal=1)

        # check that Rician corrects produces larger noise estimate
        assert_(
            np.mean(
                pca_noise_estimate(
                    data, gtab, correct_bias=True, images_as_samples=images_as_samples
                )
            )
            > np.mean(
                pca_noise_estimate(
                    data, gtab, correct_bias=False, images_as_samples=images_as_samples
                )
            )
        )

        assert_warns(UserWarning, pca_noise_estimate, data, gtab, patch_radius=0)



import numpy as np

from dipy.denoise.denspeed import nlmeans_3d
from dipy.testing.decorators import warning_for_keywords

# from warnings import warn
# import warnings

# warnings.simplefilter('always', DeprecationWarning)
# warn(DeprecationWarning("Module 'dipy.denoise.nlmeans' is deprecated,"
#                         " use module 'dipy.denoise.non_local_means' instead"))


@warning_for_keywords()
def nlmeans(
    arr,
    sigma,
    *,
    mask=None,
    patch_radius=1,
    block_radius=5,
    rician=True,
    num_threads=None,
):
    r"""Non-local means for denoising 3D and 4D images.

    See :footcite:p:Descoteaux2008a` for further details about the method.

    Parameters
    ----------
    arr : 3D or 4D ndarray
        The array to be denoised
    mask : 3D ndarray
    sigma : float or 3D array
        standard deviation of the noise estimated from the data
    patch_radius : int
        patch size is ``2 x patch_radius + 1``. Default is 1.
    block_radius : int
        block size is ``2 x block_radius + 1``. Default is 5.
    rician : boolean
        If True the noise is estimated as Rician, otherwise Gaussian noise
        is assumed.
    num_threads : int, optional
        Number of threads to be used for OpenMP parallelization. If None
        (default) the value of OMP_NUM_THREADS environment variable is used
        if it is set, otherwise all available threads are used. If < 0 the
        maximal number of threads minus $|num_threads + 1|$ is used (enter -1 to
        use as many threads as possible). 0 raises an error.

    Returns
    -------
    denoised_arr : ndarray
        the denoised ``arr`` which has the same shape as ``arr``.

    References
    ----------
    .. footbibliography::

    """

    # warn(DeprecationWarning("function 'dipy.denoise.nlmeans'"
    #                        " is deprecated, use module "
    #                        "'dipy.denoise.non_local_means'"
    #                        " instead"))

    if arr.ndim == 3:
        sigma = np.ones(arr.shape, dtype=np.float64) * sigma
        return nlmeans_3d(
            arr,
            mask=mask,
            sigma=sigma,
            patch_radius=patch_radius,
            block_radius=block_radius,
            rician=rician,
            num_threads=num_threads,
        ).astype(arr.dtype)

    elif arr.ndim == 4:
        denoised_arr = np.zeros_like(arr)

        if isinstance(sigma, np.ndarray) and sigma.ndim == 3:
            sigma = np.ones(arr.shape, dtype=np.float64) * sigma[..., np.newaxis]
        else:
            sigma = np.ones(arr.shape, dtype=np.float64) * sigma

        for i in range(arr.shape[-1]):
            denoised_arr[..., i] = nlmeans_3d(
                arr[..., i],
                mask=mask,
                sigma=sigma[..., i],
                patch_radius=patch_radius,
                block_radius=block_radius,
                rician=rician,
                num_threads=num_threads,
            ).astype(arr.dtype)

        return denoised_arr

    else:
        raise ValueError("Only 3D or 4D array are supported!", arr.shape)



import nibabel as nib
import numpy as np
import numpy.testing as npt
from scipy.ndimage import binary_erosion
from scipy.stats import pearsonr

from dipy.core.sphere import HemiSphere
from dipy.data import get_fnames, get_sphere
from dipy.direction.peaks import peaks_from_positions
from dipy.direction.pmf import SimplePmfGen, SHCoeffPmfGen
from dipy.reconst.shm import sh_to_sf
from dipy.tracking.tractogen import generate_tractogram
from dipy.tracking.stopping_criterion import BinaryStoppingCriterion
from dipy.tracking.streamline import Streamlines
from dipy.tracking.tracker_parameters import generate_tracking_parameters
from dipy.tracking.utils import (connectivity_matrix, random_seeds_from_mask,
                                 seeds_from_mask, seeds_directions_pairs)



def get_fast_tracking_performances(params, *, nbr_seeds=1000, nbr_threads=0):
    """
    Generate streamlines and return performances on the DiSCo dataset.
    """
    s = generate_disco_streamlines(params, nbr_seeds=nbr_seeds, nbr_threads=nbr_threads)
    return get_disco_performances(s)


def generate_disco_streamlines(params, *, nbr_seeds=1000, nbr_threads=0, sphere=None):
    """
    Return streamlines generated on the DiSCo dataset
    using the fast tracking module with the input tracking params
    """
    # fetch the disco data
    fnames = get_fnames(name="disco1", include_optional=True)

    # prepare ODFs
    if sphere is None:
        sphere = HemiSphere.from_sphere(get_sphere(name="repulsion724"))
    sh = nib.load(fnames[20]).get_fdata()
    fODFs = sh_to_sf(
        sh, sphere, sh_order_max=12, basis_type='tournier07', legacy=False
        )
    fODFs[fODFs<0] = 0
    pmf_gen = SimplePmfGen(np.asarray(fODFs, dtype=float), sphere)

    # seeds position and initial directions
    mask = nib.load(fnames[25]).get_fdata()
    affine = nib.load(fnames[25]).affine
    seed_mask = nib.load(fnames[34]).get_fdata()
    seed_mask = binary_erosion(seed_mask * mask, iterations=1)
    seed_positions = seeds_from_mask(seed_mask, affine, density=1)[:nbr_seeds]
    peaks = peaks_from_positions(seed_positions, fODFs, sphere, npeaks=1, affine=affine)
    seeds, directions = seeds_directions_pairs(seed_positions, peaks, max_cross=1)

    # stopping criterion
    sc = BinaryStoppingCriterion(mask)

    streamlines = generate_tractogram(seeds,
                                      directions,
                                      sc,
                                      params,
                                      pmf_gen,
                                      affine=affine,
                                      nbr_threads=nbr_threads)
    return streamlines


def get_disco_performances(streamlines):
    """
    Return the streamlines connectivity performance compared to the GT DiSCo connectom.
    """
    # fetch the disco data
    fnames = get_fnames(name="disco1", include_optional=True)

    # prepare the GT connectome data
    GT_connectome = np.loadtxt(fnames[35])
    connectome_mask = np.tril(np.ones(GT_connectome.shape), -1) > 0
    labels_img = nib.load(fnames[23])
    affine = labels_img.affine
    labels = np.round(labels_img.get_fdata()).astype(int)
    connectome = connectivity_matrix(streamlines, affine, labels)[1:,1:]

    r, _ = pearsonr(GT_connectome[connectome_mask].flatten(),
                    connectome[connectome_mask].flatten())
    return r


def test_tractogram_reproducibility():
    # Test tractogram reproducibility
    params0 = generate_tracking_parameters("prob",
                                           max_len=500,
                                           step_size=0.2,
                                           voxel_size=np.ones(3),
                                           max_angle=20,
                                           random_seed=0)

    params1 = generate_tracking_parameters("prob",
                                           max_len=500,
                                           step_size=0.2,
                                           voxel_size=np.ones(3),
                                           max_angle=20,
                                           random_seed=1)

    params2 = generate_tracking_parameters("prob",
                                           max_len=500,
                                           step_size=0.2,
                                           voxel_size=np.ones(3),
                                           max_angle=20,
                                           random_seed=2)

    # Same random generator seed
    r1 = get_fast_tracking_performances(params1, nbr_seeds=100, nbr_threads=1)
    r2 = get_fast_tracking_performances(params1, nbr_seeds=100, nbr_threads=1)
    npt.assert_equal(r1, r2)

    # Random random generator seed
    r1 = get_fast_tracking_performances(params0, nbr_seeds=100, nbr_threads=1)
    r2 = get_fast_tracking_performances(params0, nbr_seeds=100, nbr_threads=1)
    npt.assert_(not r1 == r2, msg="Tractograms are identical (random seed).")

    # Different random generator seeds
    r1 = get_fast_tracking_performances(params1, nbr_seeds=100, nbr_threads=1)
    r2 = get_fast_tracking_performances(params2, nbr_seeds=100, nbr_threads=1)
    npt.assert_(not r1 == r2, msg="Tractograms are identical (different seeds).")


def test_tracking_max_angle():
    """This tests that the angle between streamline points is always smaller
    then the input `max_angle` parameter.
    """

    def get_min_cos_similarity(streamlines):
        min_cos_sim = 1
        for sl in streamlines:
            if len(sl) > 1:
                v = sl[:-1] - sl[1:]  # vectors have norm of 1
                for i in range(len(v) - 1):
                    cos_sim = np.dot(v[i], v[i + 1])
                    if cos_sim < min_cos_sim:
                        min_cos_sim = cos_sim
        return min_cos_sim

    for sph in [
        get_sphere(name="repulsion100"),
        HemiSphere.from_sphere(get_sphere(name="repulsion100")),
    ]:
        for max_angle in [20,45]:

            params = generate_tracking_parameters("det",
                                                max_len=500,
                                                step_size=1,
                                                voxel_size=np.ones(3),
                                                max_angle=max_angle,
                                                random_seed=0)

            streamlines = generate_disco_streamlines(params, nbr_seeds=100, sphere=sph)
            min_cos_sim = get_min_cos_similarity(streamlines)
            npt.assert_(np.arccos(min_cos_sim) <= np.deg2rad(max_angle))


def test_tracking_step_size():
    """This tests that the distance between streamline
    points is equal to the input `step_size` parameter.
    """

    def get_points_distance(streamlines):
        dists = []
        for sl in streamlines:
            dists.extend(np.linalg.norm(sl[0:-1] - sl[1:], axis=1))
        return dists


    for step_size in [0.02, 0.5, 1]:
        params = generate_tracking_parameters("det",
                                              max_len=500,
                                              step_size=step_size,
                                              voxel_size=np.ones(3),
                                              max_angle=20,
                                              random_seed=0)

        streamlines = generate_disco_streamlines(params, nbr_seeds=100)
        dists = get_points_distance(streamlines)
        npt.assert_almost_equal(np.min(dists), step_size)
        npt.assert_almost_equal(np.max(dists), step_size)


def test_return_all():
    """This tests that the number of streamlines equals the number of seeds
    when return_all=True.
    """


    fnames = get_fnames(name="disco1", include_optional=True)
    sphere = HemiSphere.from_sphere(get_sphere(name="repulsion724"))
    sh = nib.load(fnames[20]).get_fdata()
    fODFs = sh_to_sf(sh, sphere, sh_order_max=12, basis_type='tournier07', legacy=False)
    fODFs[fODFs<0] = 0
    pmf_gen = SimplePmfGen(np.asarray(fODFs, dtype=float), sphere)

    # seeds position and initial directions
    mask = nib.load(fnames[25]).get_fdata()
    sc = BinaryStoppingCriterion(mask)
    affine = nib.load(fnames[25]).affine
    seed_mask = np.ones(mask.shape)
    seeds = random_seeds_from_mask(seed_mask, affine, seeds_count=100,
                                   seed_count_per_voxel=False)
    directions = np.random.random(seeds.shape)
    directions = np.array([v/np.linalg.norm(v) for v in directions])

    # test return_all=True
    params = generate_tracking_parameters("det",
                                          max_len=500,
                                          min_len=0,
                                          step_size=0.5,
                                          voxel_size=np.ones(3),
                                          max_angle=20,
                                          random_seed=0,
                                          return_all=True)
    stream_gen = generate_tractogram(seeds,
                                     directions,
                                     sc,
                                     params,
                                     pmf_gen,
                                     affine=affine)
    streamlines = Streamlines(stream_gen)
    npt.assert_equal(len(streamlines), len(seeds))

    # test return_all=False
    params = generate_tracking_parameters("det",
                                          max_len=500,
                                          min_len=10,
                                          step_size=0.5,
                                          voxel_size=np.ones(3),
                                          max_angle=20,
                                          random_seed=0,
                                          return_all=False)
    stream_gen = generate_tractogram(seeds,
                                     directions,
                                     sc,
                                     params,
                                     pmf_gen,
                                     affine=affine)
    streamlines = Streamlines(stream_gen)
    npt.assert_array_less(len(streamlines), len(seeds))


def test_max_min_length():
    """This tests that the returned streamlines respect the length criterion.
    """
    fnames = get_fnames(name="disco1", include_optional=True)
    sphere = HemiSphere.from_sphere(get_sphere(name="repulsion724"))
    sh = nib.load(fnames[20]).get_fdata()
    fODFs = sh_to_sf(
        sh, sphere, sh_order_max=12, basis_type='tournier07', legacy=False
        )
    fODFs[fODFs<0] = 0
    pmf_gen = SimplePmfGen(np.asarray(fODFs, dtype=float), sphere)

    # seeds position and initial directions
    mask = nib.load(fnames[25]).get_fdata()
    sc = BinaryStoppingCriterion(mask)
    affine = nib.load(fnames[25]).affine
    seed_mask = np.ones(mask.shape)
    seeds = random_seeds_from_mask(seed_mask, affine, seeds_count=1000,
                                   seed_count_per_voxel=False)
    directions = np.random.random(seeds.shape)
    directions = np.array([v/np.linalg.norm(v) for v in directions])

    min_len=10
    max_len=100

    params = generate_tracking_parameters("det",
                                          max_len=max_len,
                                          min_len=min_len,
                                          step_size=0.5,
                                          voxel_size=np.ones(3),
                                          max_angle=20,
                                          random_seed=0,
                                          return_all=False)
    stream_gen = generate_tractogram(seeds,
                                     directions,
                                     sc,
                                     params,
                                     pmf_gen,
                                     affine=affine)
    streamlines = Streamlines(stream_gen)
    errors = np.array([len(s) < min_len and len(s) > max_len for s in streamlines])

    npt.assert_(np.sum(errors) == 0)


def test_buffer_frac():
    """This tests that the buffer fraction for generate tractogram plays well.
    """
    fnames = get_fnames(name="disco1", include_optional=True)
    sphere = HemiSphere.from_sphere(get_sphere(name="repulsion724"))
    sh = nib.load(fnames[20]).get_fdata()
    fODFs = sh_to_sf(
        sh, sphere, sh_order_max=12, basis_type='tournier07', legacy=False
        )
    fODFs[fODFs<0] = 0
    pmf_gen = SimplePmfGen(np.asarray(fODFs, dtype=float), sphere)

    # seeds position and initial directions
    mask = nib.load(fnames[25]).get_fdata()
    sc = BinaryStoppingCriterion(mask)
    affine = nib.load(fnames[25]).affine
    seed_mask = np.ones(mask.shape)
    seeds = random_seeds_from_mask(seed_mask, affine, seeds_count=500,
                                   seed_count_per_voxel=False)
    directions = np.random.random(seeds.shape)
    directions = np.array([v/np.linalg.norm(v) for v in directions])

    params = generate_tracking_parameters("prob",
                                          max_len=500,
                                          min_len=0,
                                          step_size=0.5,
                                          voxel_size=np.ones(3),
                                          max_angle=20,
                                          random_seed=0,
                                          return_all=True)

    streams = Streamlines(generate_tractogram(seeds,
                                              directions,
                                              sc,
                                              params,
                                              pmf_gen,
                                              affine=affine,
                                              buffer_frac=1.0))

    # test the results are identical with various buffer fractions
    for frac in [0.01,0.1,0.5]:
        frac_streams = Streamlines(generate_tractogram(seeds,
                                                       directions,
                                                       sc,
                                                       params,
                                                       pmf_gen,
                                                       affine=affine,
                                                       buffer_frac=frac))
        npt.assert_equal(len(frac_streams), len(streams))
